import matplotlib.pyplot as plt
from matplotlib import gridspec
import numpy as np
from datetime import datetime
from pathlib import Path
from GRATIOSA import Genome
from GRATIOSA.globvar import *

'''
This module plots experimental information (one subplot per experiment) and 
gene annotations for a region. The genes annotation subplot is created with
the subplot_genes function. RNASeq subplots and Chipseq subplots are created 
with the subplot_rnaseq_coverage and subplot_signal functions respectively.
'''


plt.rcParams.update({'pdf.fonttype': 42})
plt.rcParams.update({'ps.fonttype': 42})
plt.rcParams.update({'font.size': 12})
plt.rcParams.update({'legend.fontsize': 10})
plt.rcParams.update({'font.family': "Arial"})


def plot_region(gen, beg, end,
                RNASeq_cond=[], 
                signals_cond=[], 
                gene_names=True,
                output_dir=f"{resdir}Genome_plot/",
                output_file=f"Genome_{datetime.now()}",
                file_extension=".pdf",
                *args, **kwargs):
    """
    Plots experimental information (one subplot per experiment) and gene
    annotations for a region. The genes annotation subplot is created with
    the subplot_genes function. RNASeq subplots and Chipseq subplots
    are created with the subplot_rnaseq_coverage and subplot_signal functions
    respectively.

    Args:
        gen: Genome instance
        beg (int.): beginning of the region to be plotted
        end (int.): end of the region to be plotted
        RNASeq_cond (list of str.): list of the RNASeq conditions names to
            represent. Left empty by default ie no cover is plotted.
        tr_object (Transcriptome instance): Transcriptome instance with 
            loaded RNASeq coverages
        signals_cond (list of str.): list of the Chipseq conditions names to
            represent. Left empty by default ie no signal is plotted.
        ch_object (Chipseq instance): Chipseq instance with loaded signals.
        gene_names (Optional [Bool.]): If true, the name of the genes is noted on their
            representation. (True by default).
        output_dir (Optional [str.]): output directory
        output_file (Optional [str.]): output filename for plot
        file_extension (Optional [str.]): Graphic file extension type (.pdf by default)
        R_ylabels (Optional [list of str.]): y-axis label for RNASeq data subplots.
            By default, conditions names given in input with the RNASeq_cond
            argument.
        S_ylabels (Optional [list of str.]): y-axis label for RNASeq data subplots.
            By default, conditions names given in input with the signals_cond
            argument.
        vlines (Optional [dict.]): Dictionnary of shape {position: annotation text}.
            One green vertical line and its annotation are plotted for each
            item of this dictionnary.
        figsize (Optional [(float,float)]): width and height in inches
            (by default: (3.5,n) with n the number of suplots)
        hratios (Optional [array-like]): Relative heights of the subplots. If not given,
            all rows will have the same height except the genes subplot.
        hspace (Optional [float]): Height space between subplots, expressed as a 
            fraction of the average axis height. By default: 0.05
        figpos (Optional [array-like of length 4]): Position of the subplots as a 
            fraction of figure width or height. The array contains the 
            positions in the following order: bottom, top, left and right.

    
    Warning:
        This method needs a genomic annotation. If no annotation is
        loaded to the Genome instance, the load_annotation method of with
        the default "sequence.gff3" file is computed. To use another 
        annotation, please load an annotation to your Genome instance before 
        using this method.

    Warning:
        To plot a ChIPseq coverage, all the selected conditions must first be loaded 
        on the Chipseq instance using one of the ChipSeq method (for example
        load_signal or load_signals_average). This Chipseq instance must
        then be given as input with the argument "ch_object".

    Warning:
        To plot a RNAseq coverage, all the selected conditions must first be loaded 
        on the Transcriptome instance using tr.load_rnaseq_cov() and this
        instance must be given as input with the argument "tr_object".

    Example:
        >>> from GRATIOSA import Genome, Transcriptome, Chipseq, plot_genome
        >>> g = Genome.Genome("ecoli")
        >>> g.load_annotation()
        >>> tr = Transcriptome.Transcriptome("ecoli")
        >>> tr.load_rnaseq_cov()
        >>> ch = Chipseq.Chipseq("ecoli")
        >>> ch.load_signals_average(list_cond=["cond1","cond2"],
        ...                         average_name="cond12")
        >>> plot_genome.plot_region(g,2101000,2106000,
                                    RNASeq_cond = ["WT"],
                                    signals_cond=["cond12"],
                                    R_ylabels=["WT"],
                                    S_ylabels=["cond12"],
                                    tr_object=tr,
                                    ch_object=ch,
                                    hratios = [1,2,2,1],
                                    figsize=(6,5),
                                    figpos=[0.1,0.98,0.15,0.95])
    """

    # loads RNASeq data
    if len(RNASeq_cond) != 0:
        try:
            tr = kwargs.get('tr_object')
        except BaseException:
            sys.exit("Please give a Transcriptome instance with loaded RNASeq coverages in input with the 'tr_object' argument.")
        error = set(RNASeq_cond) - set(tr.rnaseq_cov_pos.keys())
        if len(error) != 0:
            sys.exit(
                f"{error} not found in the RNASeq coverages. Available coverages are {list(tr.rnaseq_cov_pos.keys())}")
        if isinstance(RNASeq_cond, str):
            RNASeq_cond = [RNASeq_cond]
        R_ylabels = kwargs.get('R_ylabels', RNASeq_cond)
        if isinstance(R_ylabels, str):
            R_ylabels = [R_ylabels]

    # loads ChipSeq data
    if len(signals_cond) != 0:
        try:
            ch = kwargs.get('ch_object')
            ch.get_all_signals()
        except BaseException:
            sys.exit("Please give a Chipseq instance with loaded signal in input with the 'ch_object' argument.")
        error = set(signals_cond) - set(ch.all_signals.keys())
        if len(error) != 0:
            sys.exit(f"{error} not found in the loaded signals. Available signals are {list(ch.all_signals.keys())}")
        if isinstance(signals_cond, str):
            signals_cond = [signals_cond]
        S_ylabels = kwargs.get('S_ylabels', signals_cond)
        if isinstance(S_ylabels, str):
            S_ylabels = [S_ylabels]

    # figure settings
    l = len(RNASeq_cond) + len(signals_cond)
    if l < 4:
        height_r = [1] + [2] * l
    else:
        height_r = [1] * (l + 1)
    hratios = kwargs.get("hratios", height_r)
    hspace = kwargs.get("hspace", 0.05)
    figsize = kwargs.get("figsize", (3.5, l + 1))
    fig = plt.figure(figsize=figsize)
    figpos = kwargs.get("figpos", [0.15, 0.98, 0.25, 0.9])
    gs = gridspec.GridSpec(l + 1, 1, height_ratios=hratios, bottom=figpos[0],
                           top=figpos[1], left=figpos[2], right=figpos[3],
                           hspace=hspace)
    # xticks
    dist = end - beg
    tickda = np.array([100, 200, 500, 1000, 2000, 5000, 10000,
                      20000, 50000, 100000, 200000, 500000, 1000000])
    tickdist = tickda[np.argmin(np.abs(1 - 5 * tickda / (end - beg)))]
    xticks = np.arange(
        (beg / tickdist) * tickdist,
        (end / tickdist + 1) * tickdist,
        tickdist)
    xticks = [None] * l + [xticks]

    # genes annotation subplot
    ax = plt.subplot(gs[0])
    subplot_genes(ax, gen, beg, end, xticks=xticks[0], gene_names=gene_names)
    vlines = kwargs.get('vlines', None)
    if vlines is not None:
        for v in vlines.keys():
            ax.axvline(x=v - beg, color="green", lw=1, ls="-")
            ax.annotate(vlines[v], xy=(v - beg, 0), color="green", va='center')

    n = 1
    # RNASeq subplots
    for nr, Rcond in enumerate(RNASeq_cond):
        ax = plt.subplot(gs[n])
        subplot_rnaseq_coverage(ax, tr, Rcond, beg, end,
            ylabel=R_ylabels[nr],
            xticks=xticks[n])
        n += 1

    # ChipSeq subplots
    for ns, Scond in enumerate(signals_cond):
        ax = plt.subplot(gs[n])
        subplot_signal(ax,ch,Scond,beg,end,
            ylabel=S_ylabels[ns],
            xticks=xticks[n])
        n += 1

    fig.align_ylabels()
    ax.set_xlabel("Genomic position")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    pathf = output_dir + output_file + file_extension
    plt.savefig(pathf,transparent=False,facecolor='white')
    print(f"Saved as {pathf}")
    plt.show()
    plt.close()


def subplot_genes(ax, gen, beg, end, xticks=None, gene_names=True):
    """
    Plots a subplot with the annoted oriented genes that are located in
    the selected region.

    Args:
        ax (matplotlib Axes instance)
        gen (Genome instance)
        beg (int.): beginning of the region to be plotted
        end (int.): end of the region to be plotted
        xticks (list of float): xaxis' tick locations.
            None by default ie. the xaxis is left empty.
        gene_names (Bool.): If true, the name of the genes is noted on their
            representation. (True by default).

    Warnings:
        This method needs a genomic annotation. If no annotation is
        loaded to the Genome instance, the load_annotation method of with
        the default "sequence.gff3" file is computed. To use another 
        annotation, please load an annotation to your Genome instance before using this method.
    """
    if not hasattr(gen, "genes"):
        gen.load_annotation()

    if xticks is not None:
        ax.set_xticks(xticks)
    else:
        ax.set_xticklabels([])
        ax.set_xticks([])

    length = end - beg
    text_up = False

   # take only genes in region
    gene_list = [g for g in gen.genes.values() if g.right >
                 beg and g.left < end]

    # sort list
    gene_start = np.argsort([(g.left + g.right) / 2 for g in gene_list])
    gene_list = [gene_list[i] for i in gene_start]

    for i, g in enumerate(gene_list):

        if g.strand:
            # when arrow is too little, a rectangle is drawn instead
            if g.end - g.start < length / 20:
                hlenght = 0
            else:
                hlenght = 1
            ax.quiver(g.start, 1, g.end - g.start, 0, angles='xy',
                scale_units='xy', scale=1, width=0.05, headwidth=1, 
                headlength=hlenght, headaxislength=1, minlength=0)

            if gene_names and beg < (g.end + g.start) / 2 < end:
                if g.end - g.start > length / 10:
                    ax.text(
                        (g.start + g.end) / 2,
                        1,
                        "$\\it{%s}$" % (g.name),
                        size=8,
                        horizontalalignment='center',
                        verticalalignment='center',
                        color="white")
                else:
                    text_up = True  # if a gene name is written above the arrow, ylim will be modified
                    ax.text(
                        (g.start + g.end) / 2,
                        2.5,
                        "$\\it{%s}$" % (g.name),
                        size=8,
                        horizontalalignment='center',
                        verticalalignment='center',
                        color="black")

        else:
            if g.start - g.end < length / 20:
                hlenght = 0
            else:
                hlenght = 1
            ax.quiver(
                g.start,
                -2,
                g.end - g.start,
                0,
                angles='xy',
                scale_units='xy',
                scale=1,
                width=0.05,
                headwidth=1,
                headlength=hlenght,
                headaxislength=1,
                minlength=0)
            if gene_names and beg < (g.end + g.start) / 2 < end:
                if g.start - g.end > length / 10:
                    ax.text((g.start + g.end) / 2, -2, "$\\it{%s}$" %
                            (g.name), size=8, horizontalalignment='center', 
                             verticalalignment='center', color="w")
                else:
                    ax.text((g.start + g.end) / 2, -0.5, "$\\it{%s}$" %
                            (g.name), size=8, horizontalalignment='center', 
                            verticalalignment='center', color="black")

    ax.set_yticks([])
    ax.set_yticklabels([])
    ax.set_xlim(beg, end)
    if text_up:
        ax.set_ylim(-3.5, 3.5)
    else:
        ax.set_ylim(-3.5, 2.5)


def subplot_rnaseq_coverage(ax, tr, cond, beg, end, ylabel=None, xticks=None):
    """
    Plots a subplot with the with the experimental data (RNASeq coverages on
    both DNA strands) of the selected region.

    Args:
        ax (matplotlib Axes instance)
        tr (Transcriptome instance): Transcriptome instance with loaded
            RNASeq coverages
        cond (str.): RNASeq condition name to represent. 
            WARNING: This condition must first be loaded on the Transcriptome 
            instance using tr.load_rnaseq_cov().
        beg (int.): beginning of the region to be plotted
        end (int.): end of the region to be plotted
        ylabel (str.): y-axis label. By default, the condition name given
            in input with the cond argument.
        xticks (list of float): xaxis' tick locations.
            None by default ie. the xaxis is left empty.
    """
    cov_pos = tr.rnaseq_cov_pos[cond]
    cov_neg = tr.rnaseq_cov_neg[cond]

    x = np.arange(beg, end)
    ax.fill_between(x, cov_pos[beg:end], color="blue")
    ax.fill_between(x, -cov_neg[beg:end], color="red")
    ax.axhline(y=0, color="black", lw=1)

    if xticks is not None:
        ax.set_xticks(xticks)
    else:
        ax.set_xticklabels([])
        ax.set_xticks([])
    ax.set_xlim(beg, end)
    if ylabel is None:
        ylabel = cond
    ax.set_ylabel(ylabel)


def subplot_signal(ax, ch, cond, beg, end, ylabel=None, xticks=None):
    """
    Plots a subplot with the with the experimental data (Chipseq signal) of 
    the selected region.

    Args:
        ax (matplotlib Axes instance)
        ch (Chipseq instance): Chipseq instance with loaded signals.
        cond (str.): Chipseq condition name to represent. WARNING: This
            condition must first be loaded on the Chipseq instance using one 
            of the ChipSeq method (for example load_signal or 
            load_signals_average).
        beg (int.): beginning of the region to be plotted
        end (int.): end of the region to be plotted
        ylabel (str.): y-axis label. By default, the condition name given
            in input with the cond argument.
        xticks (list of float): xaxis' tick locations.
            None by default ie. the xaxis is left empty.
    """
    if not hasattr(ch, "all_signals"):
        ch.get_all_signals()
    coverage = ch.all_signals[cond]

    y = coverage[beg:end]
    x = np.arange(beg, end)

    test_neg = [i <= 0 for i in y]
    test_pos = [i > 0 for i in y]
    ax.fill_between(x, y, where=test_pos, color='b')
    ax.fill_between(x, y, where=test_neg, color='r')
    ax.axhline(y=0, color="black", lw=1)

    if xticks is not None:
        ax.set_xticks(xticks)
    else:
        ax.set_xticklabels([])
        ax.set_xticks([])
    ax.set_xlim(beg, end)
    if ylabel is None:
        ylabel = cond
    ax.set_ylabel(ylabel)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

class Gene:
    '''
    The Gene class allows to gather information at the level of a single gene:
    its identifiers, its coordinates, its expression levels under different
    conditions, its functional annotations...
    '''
    def __init__(self, locus_tag, name, ID, left, right, strand, ASAP_name):
        self.locus_tag = locus_tag
        self.ID = ID
        self.name = name
        self.left = int(left)
        self.right = int(right)
        self.strand = strand
        self.middle = float(left + right) / 2
        self.length = int(right) - int(left) + 1
        if self.strand:
            self.start = int(left)
            self.end = int(right)
        else:
            self.start = int(right)
            self.end = int(left)
        self.ASAP = ASAP_name

    def add_single_rpkm(self, condition, expression_value, factor):
        if not hasattr(self, 'rpkm'):
            self.rpkm = {}

    def add_id_TSS(self, id_TSS):
        if not hasattr(self, 'id_TSS'):
            self.id_TSS = []
        if not (id_TSS in self.id_TSS):
            self.id_TSS.append(id_TSS)

    def add_fc_pval_cond(self, fc_value, condition, p_value):
        if not hasattr(self, 'fc_pval'):
            self.fc_pval = {}
        self.fc_pval[condition] = (fc_value, p_value)

    def add_state_cond(self, cond, state):
        if not hasattr(self, 'state'):
            self.state = {}
        self.state[cond] = state

    def add_left_neighbor(self, neighbor):
        self.left_neighbor = neighbor

    def add_right_neighbor(self, neighbor):
        self.right_neighbor = neighbor

    def add_orientation(self, orient):
        self.orientation = orient

    def add_expression(self, condition, expression_value):
        if not hasattr(self, 'expression'):
            self.expression = {}
        self.expression[condition] = expression_value

    def set_mean_expression(self, expression_value=None):
        if expression_value:
            self.mean_expression = expression_value
        else:
            self.mean_expression = np.mean(self.expression.values())

    def add_signal(self, cond, signal):
        if not hasattr(self, "signal"):
            self.signal = {}
        self.signal[cond] = signal

    def add_is_loop(self, cond, is_loop):
        if not hasattr(self, "is_loop"):
            self.is_loop = {}
        self.is_loop[cond] = is_loop

    def add_is_border(self, cond, is_border):
        if not hasattr(self, "is_border"):
            self.is_border = {}
        self.is_border[cond] = is_border

    def add_is_peak(self, cond, is_peak):
        if not hasattr(self, "is_peak"):
            self.is_peak = {}
        self.is_peak[cond] = is_peak

    def add_GO(self, annot, GO):
        if not hasattr(self, "GO"):
            self.GO = {}
        self.GO[annot] = GO


#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Functions called by Chipseq methods
"""

import numpy as np
from scipy import stats

def binning(data, binsize, stat="mean"):
    """
    Data binning at a chosen binsize.

    Args:
        data (array-like): A sequence of values to be binned.
        binsize (int.): The bin size
        stat (str.): The statistic to compute ('mean' by default). The 
                following statistics are available: mean, std, median, count, 
                sum, min, max. See scipy.stats.binned_statistic documentation 
                for more details)

    Returns:
        Array containing the values of the selected statistic in each bin.
    """
    # determines the bin coordinates
    l_data = len(data)
    bins = np.arange(0, l_data, binsize)
    bins = np.append(bins, l_data)

    # performs the binning
    binned_data = stats.binned_statistic(
        np.arange(l_data), data, bins=bins, statistic=stat)

    return binned_data


def smoothing(data, window):
    """
    Data smoothing by moving average of CIRCULAR data.

    Args:
        data (array-like): A sequence of values to be smoothed.
        window (int.): window size. The value of the smoothed signal of a 
                position p is equal to the average of the signal between 
                p - window/2 and p + window/2.

    Returns:
        Array containing the smoothed data.
    """
    s_data = []
    i = 0
    w = int(window / 2)
    # circular DNA
    while i < w:
        window_average = np.mean(list(data[i - w:]) + list(data[0:i + w + 1]))
        s_data.append(window_average)
        i += 1
    while i < len(data) - w:
        window_average = np.mean(data[i - w:i + w + 1])
        s_data.append(window_average)
        i += 1
    while i < len(data):
        window_average = np.mean(
            list(data[i - w:]) + list(data[0:i + w - len(data) + 1]))
        s_data.append(window_average)
        i += 1
    return s_data


def load_sites_cond(path2file, startline, sep, start_col, end_col, val_col=None):
    '''
    Called by load_peaks, allows the peaks data to be loaded by specifying
    files (typically a .BED file of peaks obtained with MACS2), and where 
    each information is (one column for each type of information).

    Args:
        path2file (str.): path to the file containing the data
        startline (int.): file start line
        sep (str.): file separator
        start_col (int.): index of the column containing the position of the
                beginning of the peak
        end_col (int.): index of the column containing the position of the 
                end of the peak
        val_col (int.): index of the column containing the position of the 
                value associated to the peak (default: None)

    Returns:
        List of tuples of shape (start,end). One tuple represents one peak.

    Note: 
        Column numbering starts at 0.
    '''
    peaks = {}
    if sep == '\\t':
        sep = '\t'

    with open(path2file, 'r') as f:
        i = 0
        while i < startline:
            header = next(f)
            i += 1
        for line in f:
            line = line.strip('\n').split(sep)
            try:
                if val_col != None :
                    peaks[(int(line[start_col]), int(line[end_col]))] = float(line[val_col])
                else : 
                    peaks[(int(line[start_col]), int(line[end_col]))] = None
            except Exception as e:
                print(e)
    f.close()

    return peaks


#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Functions called by HiC methods
"""


def load_HiC_site_cond(site_type, 
                       path2file, 
                       startline, 
                       sep, 
                       bin1_col,
                       binsize, 
                       bin2_col=None, 
                       score_col=None, 
                       pval_col=None, 
                       qval_col=None):
    '''
    Called by load_HiC_borders and load_HiC_loops, allows borders and loops 
    data to be loaded by specifying files, and where each information is (one 
    column for each type of information).

    Args:
        type (str.): HiC detected motif type ("borders" or "loops")
        path2file (str.): path to the file containing the data
        startline (int.): file start line
        sep (str.): file separator
        bin1_col (int.): index of the column containing the position of the 
                first bin
        binsize (int.): binsize (in b)
        bin2_col (Optional [int.]): index of the column containing the position of the
                second bin
        score_col (Optional [int.]): index of the column containing the position of the score
        pval_col (Optional [int.]): index of the column containing the position of the pvalue
        qval_col (Optional [int.]): index of the column containing the position of the qvalue

    Note: 
        Column numbering starts at 0.
    '''
    if sep == '\\t':
        sep = '\t'

    if site_type == "borders":
        Borders = {}
    elif site_type == "loops":
        Loops = {}

    with open(path2file, 'r') as f:
        i = 0
        while i < startline:
            header = next(f)
            i += 1
        for line in f:
            line = line.strip('\n').split(sep)
            try:
                if site_type.lower() == "borders":
                    b = int(line[int(bin1_col)]) * binsize
                    Borders[b] = {"binsize": binsize}
                    if score_col:
                        Borders[b]['score'] = line[int(score_col)]
                    if pval_col:
                        Borders[b]['pval'] = line[int(pval_col)]
                    if qval_col:
                        Borders[b]['qval'] = line[int(qval_col)]
                elif site_type.lower() == "loops":
                    pos_tuple = (
                        int(line[int(bin1_col)]) * binsize, int(line[int(bin2_col)]) * binsize)
                    Loops[pos_tuple] = {"binsize": binsize}
                    if score_col:
                        Loops[pos_tuple]['score'] = line[int(score_col)]
                    if pval_col:
                        Loops[pos_tuple]['pval'] = line[int(pval_col)]
                    if qval_col:
                        Loops[pos_tuple]['qval'] = line[int(qval_col)]
            except Exception as e:
                print(e)
    f.close()

    if site_type == "borders":
        return Borders

    elif site_type == "loops":
        return Loops



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Functions called by Genome methods
"""
import numpy as np
from GRATIOSA.globvar import *
from GRATIOSA.Gene import Gene
from GRATIOSA.TSS_TTS_TU import TSS, TTS, TU


def read_seq(filename):
    '''
    Called by load_seq, read_seq allows genomic sequence to be loaded from a
    .fasta file

    Args:
        filename (str.): name of the file containing the DNA sequence in
                         FASTA format.

    Returns:
        genomic sequence (str.)

    '''
    seq = str()
    seq_file = open(filename, "r")
    for i in seq_file.readlines():
        line = i.strip()  # Removes \n
        if line != '':  # Inspect if empty line
            if line[0] != ">":
                seq += line
    seq_file.close
    return seq


def load_gff(annotations_filename):
    '''
    Called by load annotation, load_gff allows genes to be loaded from .gff 
    or .gff3 annotation file. For each gene of the annotation, onne Gene 
    object is created and intialized with the following attributes: 
    locus_tag, ID, name, left (left coordinate), right (right coordinate), 
    middle, length, strand, start, end and ASAP.

    Args:
        filename (str.): name of the annotation file.

    Returns:
        Dictionary: Dictionary of shape {locus: Gene object} with each Gene 
        object initialized with the following attributes:

            * locus_tag: "locus_tag" recorded in the annotation,
            * ID: "ID" or "locus_tag" if no "ID" is associated this gene in 
              the annotation,
            * name: "gene" or "locus_tag" if no "gene" is associated to this
              gene in the annotation,
            * ASAP: "ASAP" name (By default: '').
            * strand: gene strand,
            * left, right: gene coordinates (does not take into account the
              strand, ie right > left)
            * start, end, middle: positions of the beginning, the middle and
              the end of the gene
            * length: gene length (=right-left)


    Warnings: 
        ASAP name  is found only if it is noted on the same line as the
        "name" line or the line next to it
    '''
    genes_dict = {}
    annot_file = open(annotations_filename, "r")
    for line in annot_file.readlines():
        ASAP_name = ''
        if line[0] != '#':
            if line != '\n':
                line = line.split('\t')
                gene_info = {}
                for x in line[8].split(';'):
                    x = x.strip().split('=')
                    gene_info[x[0]] = x[1]
                if ('locus_tag' in gene_info):
                    locus = gene_info['locus_tag']
                    if "ID" in gene_info.keys():
                        ID = gene_info["ID"]
                    else:
                        ID = locus
                    if "gene" in gene_info.keys():
                        name = gene_info["gene"]
                    else:
                        name = locus

                    if "Dbxref" in gene_info.keys():
                        db_ref = "Dbxref"
                    elif "db_xref" in gene_info.keys():
                        db_ref = "db_xref"
                    else:
                        db_ref = None
                    if db_ref is not None:
                        dbxref = gene_info[db_ref].split(",")
                        for db in dbxref:
                            if "ASAP" in db:
                                ASAP_name = db.split(":")[1]
                    left, right = map(int, line[3:5])
                    strand = True if line[6] == "+" else False
                    genes_dict[locus] = Gene(
                        locus, name, ID, left, right, strand, ASAP_name)

                elif line[2] == "CDS" and ASAP_name == '':
                    tleft, tright = map(int, line[3:5])
                    if tleft == left and tright == right:
                        if "Dbxref" in gene_info.keys():
                            db_ref = "Dbxref"
                        elif "db_xref" in gene_info.keys():
                            db_ref = "db_xref"
                        else:
                            db_ref = None
                        if db_ref is not None:
                            dbxref = gene_info[db_ref].split(",")
                            for db in dbxref:
                                if "ASAP" in db:
                                    ASAP_name = db.split(":")[1]
                        genes_dict[locus] = Gene(
                            locus, name, ID, left, right, strand, ASAP_name)
    return genes_dict


def load_annot_general(annotations_filename, 
                       separator, 
                       tag_column, 
                       name_column,
                       ID_column, 
                       strand_column, 
                       left_column, 
                       right_column, 
                       ASAP_column, 
                       start_line):
    '''
    Called by load annotation, load_annot_general allows genes to be loaded
    from any annotation file, although most of the time, annotation is loaded
    from gff files (using the load_gff function). The annotation file has to
    contains one row per genes and the following columns: locus_tag, ID, 
    name, ASAP, left coordinate, right coordinate and strand.
    For each gene of the annotation, one Gene object is created and 
    intialized with the following attributes: locus_tag, ID, name, left 
    (left coordinate), right (right coordinate), middle, length, strand, 
    start, end and ASAP.

    Args:
        filename (str.): name of the annotation file
        separator (str.): file separator
        tag_col (int.): index of the column containing the locus tags 
        name_column (int.):index of the column containing the names 
        ID_column (int.): index of the column containing the ID
        strand_column (int.): index of the column containing the strand 
        left_column (int.): index of the column containing the left 
                coordinates
        right_column (int.): index of the column containing the right 
                coordinates
        ASAP_column (int.):index of the column containing the ASAP names
        start_line (int.): file start line
    
    Returns:
        Dictionary: Dict. of shape {locus: Gene object} with each Gene object
        initialized with the following attributes: locus_tag, ID, name,
        left (left coordinate), right (right coordinate), middle, length, 
        strand, start, end and ASAP.

    Note: 
        Column numbering starts at 0.

        Strands can be noted with one of the following writings:

            * Forward: "forward","1","+","true","plus" (case insensitive)
            * Reverse: "complement","-1","-","false","minus" (case insensitive)
    '''
    genes_dict = {}
    with open(annotations_filename, 'r') as f:
        head = next(f)
        headl = head.strip()
        if separator == '\\t':
            separator = '\t'
        headl = headl.split(separator)
        i = 0
        while i < start_line:
            header = next(f)
            i += 1
        for line in f:
            try:
                line = line.strip()
                line = line.split(separator)
                locus = line[tag_column]
                name = line[name_column]
                ID = line[ID_column]
                ASAP = line[ASAP_column]
                if line[strand_column].lower() in ["complement", "-1", "-", "false", "minus"]:
                    strand = False
                elif line[strand_column].lower() in ["forward", "1", "+", "true", "plus"]:
                    strand = True
                else:
                    print("Unrecognized strand type")
                    sys.exit("Unrecognized strand type")
                left = line[left_column]
                right = line[right_column]
                genes_dict[locus] = Gene(
                    locus, name, ID, left, right, strand, ASAP)
            except Exception as e:
                if line != ['']:
                    print(f"WARNING: Error with {line}")
    return genes_dict


def load_TSS_cond(genes_dict, 
                  filename, 
                  TSS_column, 
                  start_line, 
                  separator,
                  strandcol, 
                  genescol=None, 
                  sigcol=None, 
                  sitescol=None, 
                  scorecol=None,):
    '''
    Called by load_TSS, load_TSS_cond allows TSS data to be loaded from any 
    file with one row per TSS and at least the following columns (separated 
    by a separator that is not a comma): TSS position and DNA strand. 
    Additional columns can be locus tags, sigma factors, sites positions and 
    score. The sites positions columns is divided in 4 part separated by 
    commas: -10 element left coordinate, -10 element right coordinate, 
    -35 element left coordinate and -35 element right coordinate.
    For each TSS of the annotation, one TSS object is created and intialized 
    with the attributes available in the data file (at least pos and strand).

    Args:
        genes_dict (dict.): dictionary of shape {locus tag: Gene object}
        filename (str.): name of the annotation file
        TSS_column (int.): index of the column containing the TSS positions
        start_line (int.): file start line
        separator (str.): file separator (other than commas !)
        strandcol (int.): index of the column containing the strands in the 
                file
        genescol (int.): index of the column containing the locus tags of the
                genes associated to each TSS in the file. By default: None 
                (ie not in the file)
        sigcol (int.): index of the column containing the name of the sigma 
                factor associated to each TSS in the file
                By default: None (ie not on file)
        sitescol (int.): index of the column containing the sites positions
                         (ie the coordinates of the -10 and the -35 elements)
                         associated to each TSS in the file
                         By default: None (ie not on file)
        scorecol (int.): index of the TSS scores column in the file
                         By default: None (ie not on file)


    Returns:
        Dictionary: dict. of shape {TSS position: TSS object} with each TSS object
        initialized with, at least, the following attributes: pos and strand.
        Depending on the data present in the file and passed as arguments,
        the following attributes may also have been added: genes, score and
        promoter. The promoter attribute is a dictionary containing, for each
        sigma factor (keys) a subdictionary (value). The first created key of
        this subdictionary is "sites" and the associated value is a tuple
        containing the positions of promoter elements (-10l,-10r,-35l,-35r)
        with l = left coordinate and r = right coordinate.

    Note: 
        Column numbering starts at 0.

        Strands can be noted with one of the following writings:
        
            * Forward: "forward","1","+","true","plus" (case insensitive)
            * Reverse: "complement","-1","-","false","minus" (case insensitive)
    '''
    TSS_dict = {}  # dict of TSS objects
    with open(filename, 'r') as f:
        i = 0
        while i < start_line:
            header = next(f)
            i += 1
        for line in f:
            line = line.strip('\n')
            if separator == '\\t':
                line = line.split('\t')
            else:
                line = line.split(separator)
            try:
                pos = int(line[TSS_column])
                if line[strandcol].lower() in ["complement", "-1",
                                               "-", "false", "minus"]:
                    strand = False
                elif line[strandcol].lower() in ["forward", "1", "+", "true", "plus"]:
                    strand = True
                else:
                    strand = None
                genes = line[genescol] if genescol is not None else []
                sig = line[sigcol] if sigcol is not None else None
                sites = line[sitescol] if sitescol is not None else None
                score = line[scorecol] if scorecol is not None else None

                # if TSS needs to be init
                if pos not in TSS_dict.keys():
                    # init object tss
                    TSS_dict[pos] = TSS(pos=pos, strand=strand, score=score)
                    if genes != []:
                        TSS_dict[pos].add_genes(genes, genes_dict)
                        # add TSS to gene attributes
                        for gene in TSS_dict[pos].genes:
                            genes_dict[gene].add_id_TSS(pos)

                # Add sigma factor and binding sites to the promoter dict
                if sig is not None:  # if sigma column
                    if sites is not None:
                        TSS_dict[pos].add_promoter(sig, sites=sites)
                    else:
                        TSS_dict[pos].add_promoter(sig)

            except Exception as e:
                print('Error in line, wrong information type:', e)

    return TSS_dict


def load_TTS_cond(filename, 
                  separator, 
                  start_line, 
                  leftcol, 
                  rightcol, 
                  strandcol,
                  rhocol, 
                  seqcol=None, 
                  scorecol=None, 
                  genescol=None, 
                  *args, **kwargs):
    '''
    Called by load_TTS, load_TTS_cond allows TTS data to be loaded from any 
    file with one row per TTS and at least the following columns:
    left coordinate, right coordinate, strand and rho_dpdt. Additional 
    columns can be genes locus tags, TTS sequence and TTS score.
    For each TTS of the annotation, one TTS object is created and intialized 
    with the attributes available in the data file (at least left, right, 
    rho_dpdt and strand).

    Args:
        filename (str.): name of the annotation file
        separator (str.): file separator
        start_line (int.): file start line
        leftcol (int.): index of the column containing the TTS left 
                coordinates
        rightcol (int.): index of the column containing the TTS right 
                coordinates
        strandcol (int.): index of the column containing the TTS strands
        rhocol (int.): index of the column containing the information about
                the rho dependency of the TTS
        seqcol (int.): index of the column containing the TTS sequences
                By default: None (ie not on file)
        scorecol (int.): index of the TTS scores column in the file
                By default: None (ie not on file)
        genescol (int.): index of the column containing the locus tags of the
                genes associated to each TTS in the file.
                By default: None (ie not on file)

    Returns:
        Dictionary: Dict. of shape {TTS position: TTS object} with each TTS object
        initialized with, at least, the following attributes: left,
        right, start, end, strand and rho_dpdt. Depending on the data 
        present in the file and passed as arguments, the following attributes 
        may also have been added: genes, score and seq.

    Note: 
        Column numbering starts at 0.

        Strands can be noted with one of the following writings:
        
            * Forward: "forward","1","+","true","plus" (case insensitive)
            * Reverse: "complement","-1","-","false","minus" (case insensitive)

        Rho dependent TTS can be noted with one of the following writings: 
        "True","1" (case insensitive)
    '''

    TTS_dict = {}  # dict of TTS objects
    with open(filename, 'r') as f:
        i = 0
        while i < start_line:
            header = next(f)
            i += 1
        for line in f:
            line = line.strip('\n')
            if separator == '\\t':
                line = line.split('\t')
            else:
                line = line.split(separator)
            try:
                left = int(line[leftcol])
                right = int(line[rightcol])
                if line[strandcol].lower() in ["complement", "-1", "-", "false", "minus"]:
                    strand = False
                elif line[strandcol].lower() in ["forward", "1", "+", "true", "plus"]:
                    strand = True
                else:
                    strand = None
                rho_dpdt = True if line[rhocol].lower() in [
                    "true", "1"] else False
                seq = line[seqcol] if seqcol is not None else ""
                score = line[scorecol] if scorecol is not None else None
                genes = line[genescol] if genescol is not None else []

                newTTS = TTS(
                    left=left,
                    right=right,
                    strand=strand,
                    rho_dpdt=rho_dpdt,
                    seq=seq,
                    score=score,
                    genes=genes)
                TTS_dict[newTTS.start] = newTTS

            except Exception as e:
                print('Error in line, wrong information type:', e)

    return TTS_dict


def load_TU_cond(filename, 
                 IDcol,
                 startcol, 
                 endcol, 
                 strandcol,
                 genescol, 
                 startline, 
                 separator,
                 exprcol):
    '''
    Called by load_TU, load_TU_cond allows TU data to be loaded from any file
    with one row per TU and the following columns: start position, end 
    position, strand and genes locus tags. For each TU of the annotation, one 
    TU object is created and intialized with the attributes start, end, 
    strand, genes and expression.

    Args:
        filename (str.): name of the annotation file
        IDcol (int.): index of the column containing the TU ID
        startcol (int.): index of the column containing the TU start positions
        endcol (int.): index of the column containing the TU end positions
        strandcol (int.): index of the column containing the TU strands
        genescol (int.): index of the column containing the locus tags of the
                genes associated to each TU in the file.
                By default: None (ie not on file)
        start_line (int.): file start line
        separator (str.): file separator
        exprcol (int.): index of the column containing the TU expression

    Returns:
        Dictionary: dict. of shape {TU start: TU object} with each TU object
        initialized with the attributes start, end, strand and genes.

    Note: 
        Column numbering starts at 0.

        Strands can be noted with one of the following writings:
        
            * Forward: "forward","1","+","true","plus" (case insensitive)
            * Reverse: "complement","-1","-","false","minus" (case insensitive)
    '''

    TUs = {}
    with open(filename, 'r') as f:
        i = 0
        while i < startline:
            header = next(f)
            i += 1
        for line in f:
            line = line.strip('\n')
            if separator == '\\t':
                line = line.split('\t')
            else:
                line = line.split(separator)
            try:
                if line[strandcol].lower() in ["complement", "-1", "-", "false", "minus"]:
                    strand = False
                elif line[strandcol].lower() in ["forward", "1", "+", "true", "plus"]:
                    strand = True
                else:
                    strand = None
                if exprcol != None:
                    expr = float(line[exprcol])
                else :
                    expr = None
                TUs[int(line[IDcol])] = TU(start=int(line[startcol]), 
                                            end=int(line[endcol]), 
                                            strand=strand, 
                                            genes=line[genescol].split(","),
                                            expression=expr)
            except Exception as e:
                print(e)
    f.close()
    return TUs


#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module allows to graphically represent (barplots) the statistical
analysis performed with the module stat_analysis.
"""

import matplotlib.pyplot as plt
from datetime import datetime
import numpy as np
from GRATIOSA.globvar import *
from GRATIOSA import stat_analysis

plt.rcParams.update({'font.size': 9,
                     'font.family': "Arial",
                     'xtick.labelsize': 8,
                     'ytick.labelsize': 8,
                     'axes.labelsize': 9,
                     'axes.titlesize': 10})


def significance(pval):
    """
    Converts p-values in stars annotation.
    """
    if pval <= 0.001:
        s = '***'
    elif pval <= 0.01:
        s = '**'
    elif pval <= 0.05:
        s = '*'
    else:
        s = 'ns'
    return s


def barplot_annotate_brackets(categories, y_up, dict_pval, *args, **kwargs):
    """
    Annotates barplot with p-values, using the significance function to 
    convert p-values in stars annotation.

    Args:
        categories (list): list of the plotted categories, in the order of
                their position on the plot
        y_up (list): list of the maximal y position of each bar (taking into
                account the confidence intervals)
        dict_pval (dict.): dictionnary of shape {"cat1-cat2":pvalue} with 
                cat1 and cat2 contained in the categories list given as
                argument
        linewidth (Optional [float.]): Linewidth of the brackets. (default: 1.5)
    """
    # gets the position of each category on the barplot
    pos = {}
    for p, cat in enumerate(categories):
        pos[cat] = p

    # creates a dictionnary containing, for each significant pvalue:
    # - its pvalue converted as "*" to represent it on the plot
    # - start and end position of the bracket (positions of the 2 categories)
    # - the minimal height of the bracket
    dict_br = {}
    for cats in list(dict_pval.keys()):
        if dict_pval[cats] <= 0.05:
            s = significance(dict_pval[cats])
            c0, c1 = cats
            x0, x1 = np.sort([pos[c0], pos[c1]])
            hmax = 0
            for x in np.arange(x0, x1 + 1):
                hmax = max(hmax, y_up[x])
            dict_br[cats] = {"x0": x0, "x1": x1, "hmax": hmax, "s": s}

    # gets the ylim to defines the necessary difference between the heights
    # of 2 brackets
    ax_y0, ax_y1 = plt.gca().get_ylim()
    dh = (ax_y1 - ax_y0) / 10

    # plots the brackets without overlay
    h = 0
    linewidth = kwargs.get("linewidth", 1.5)
    while dict_br:
        h += dh
        pre_x1 = -1
        plotted_b = []
        for b in dict_br.keys():
            b_x0 = dict_br[b]["x0"]
            b_x1 = dict_br[b]["x1"]
            if b_x0 > pre_x1 and dict_br[b]["hmax"] < h:
                barx = [b_x0, b_x0, b_x1, b_x1]
                bary = [h, h + dh / 3, h + dh / 3, h]
                plt.plot(barx, bary, c='black', linewidth=linewidth)
                plt.text(float(b_x0 + b_x1) / 2, h + dh / 3, dict_br[b]["s"],
                         ha='center')
                pre_x1 = b_x1
                plotted_b.append(b)
        for b in plotted_b:
            del dict_br[b]


def plot_proportion_test(dict_cats,
                         dict_features,
                         targ_features,
                         all_features="all",
                         cats="all",
                         alt_hyp="one-sided",
                         output_dir=f"{resdir}proportion_test/",
                         output_file=f"prop_test{datetime.now()}",
                         file_extension=".pdf",
                         xlabel="",
                         ylabel="Proportion",
                         title="",
                         annot_brackets=True,
                         *args, **kwargs):
    '''
    Barplots of the proportion test: targ_features/all_features between
    categories. The proportion test, based on normal test, is computed with
    stat_analysis.proportion_test (see its documentation for more details)

    Args:
        dict_cats (dict.): classification of each elements in a dictionary of
                shape {category:[elements]}
                Example: {"border":["GeneA","GeneB"],"None":["GeneC","GeneD"]}
        dict_features (dict.): feature corresponding to each element in a
                dictionary of shape {feature:[elements]}
                Example: {"act": ["GeneA","GeneC","GeneD"], "rep":["GeneB"]}
        targ_features (list or string): targeted feature(s)
        all_features (Optional [list or string]): features that will be used to calculate
                the proportion, including targ_features default: all keys of 
                the dict_features dictionary)
        cats (Optional [list]): list of categories to compare (default: all keys of 
                dict_cats)
        alt_hyp (Optional ["two-sided" or "one-sided"]): alternative hypothesis.
                If "one-sided" is chosen, both one sided tests will be 
                perform with the statsmodels.stats.proportions_ztest 
                function and the smaller pvalue will be kept. 
                See statsmodels documentation for more details. 
                (default: "one-sided")
        output_dir (Optional [str.]): output directory
        output_file (Optional [str.]): output filename for the proportion test data 
                (.txt) and the plot
        file_extension (Optional [str.]): Graphic file extension type (.pdf by default)
        xlabel (Optional [str.]): label for the x-axis (default: empty)
        ylabel (Optional [str.]): label for the y-axis (default: "Proportion")
        title (Optional [str.]): general title for the figure
        annot_brackets (Optional [bool.]): if True, the barplot will be annotated according
                to the p-values using stars annotaion (default: True)
        brackets_linewidth (Optional [float.]): Linewidth of the brackets. (default: 1.5)
        ymin (Optional [float]): y-axis bottom limit
        ymax (Optional [float]): y-axis top limit
        figsize (Optional [(float,float)]): width and height in inches (by default: (w,2.2)
                with w dependent on the number of categories)
        xticks_rotation (Optional [int.]): x-ticks labels rotation in degrees
        xticks_labels (Optional [list.]): x-ticks labels (by default: cats)
        err_capsize (Optional [float.]): Length of the error bar caps in points
        bar_linewidth (Optional [float.]): Width of the bars edge. (default: 1.5)
        bar_width (Optional [float.]): Width of the bars. (default dependent on the number 
                of categories. If less than 5 cats: 0.7)


    Example:
        >>> import numpy as np
        >>> from GRATIOSA import plot_stat_analysis
        >>> dict_cat = {"cat1":np.arange(100,154),"cat2":np.arange(1,100),
        ...             "cat3":np.arange(154,180),"cat4":np.arange(180,230)}
        >>> dict_features ={"act":list(np.arange(1,90))+list(np.arange(100,120))
        ...                +list(np.arange(154,158))+list(np.arange(180,220)),
        ...        "rep":list(np.arange(90,94))+list(np.arange(120,150))
        ...                +list(np.arange(158,177))+list(np.arange(220,225)),
        ...        "None":list(np.arange(94,100))+list(np.arange(150,154))
        ...                +list(np.arange(177,180))+list(np.arange(225,230))}
        >>> plot_stat_analysis.plot_proportion_test(dict_cat,dict_features,"act",
        ...                all_features=["act","rep"],alt_hyp="two-sided",output_file="test")
    '''
    res = stat_analysis.proportion_test(
        dict_cats=dict_cats, dict_features=dict_features, targ_features=targ_features, all_features=all_features,
        cats=cats, alt_hyp=alt_hyp, output_dir=output_dir, output_file=output_file)

    cats = res["categories"]
    prop = res["proportions"]

    ci = np.zeros((2, len(cats)))
    for c in np.arange(len(cats)):
        ci[0, c] = prop[c] - res["confidence intervals"][c][0]
        ci[1, c] = res["confidence intervals"][c][1] - prop[c]

    if len(cats) < 5:
        wb = 0.7
    elif len(cats) < 15:
        wb = 0.5
    else:
        wb = 0.4
    figsize = kwargs.get("figsize", (len(cats) * wb/1.5+0.2, 2.2))
    plt.figure(figsize=figsize)

    bar_width = kwargs.get("bar_width", wb)
    bar_linewidth = kwargs.get("bar_linewidth", 1.5)
    err_capsize = kwargs.get("err_capsize", bar_width*5)
    plt.bar(cats, prop, yerr=ci,
            color='white', edgecolor='black', ecolor='black',
            width=bar_width, linewidth=bar_linewidth, capsize=err_capsize)

    error_max = np.amax(res['confidence intervals'], axis=1)

    if annot_brackets:
        brackets_linewidth = kwargs.get("brackets_linewidth", 1.5)
        barplot_annotate_brackets(
            cats,
            error_max,
            res["p-values"],
            linewidth=brackets_linewidth)

    plt.ylabel(ylabel)
    if xlabel != "":
        plt.xlabel(xlabel)
    ymin = kwargs.get('ymin', plt.ylim()[0])
    ymax = kwargs.get('ymax', plt.ylim()[1])
    xticks_rotation = kwargs.get("xticks_rotation", 0)
    xticks_labels = kwargs.get("xticks_labels", cats)
    plt.xticks(ticks=np.arange(len(cats)),
               labels=xticks_labels,
               rotation=xticks_rotation)
    plt.ylim(ymin, ymax)
    plt.title(title)
    plt.tight_layout()
    plt.savefig(
        f"{output_dir}{output_file}{file_extension}",
        transparent=False,
        facecolor='white')
    plt.show()
    plt.close()


def plot_enrichment_test(dict_cats,
                         dict_features,
                         targ_features,
                         all_features="all",
                         targ_cats="all",
                         all_cats="all",
                         min_nb_elements=4,
                         output_dir=f"{resdir}enrichment_test/",
                         output_file=f"enrich_test{datetime.now()}",
                         file_extension=".pdf",
                         xlabel="",
                         ylabel="Proportion",
                         title="",
                         legend_text="Global\nproportion",
                         legend_loc="best",
                         annot_star=True,
                         *args, **kwargs):
    '''
    Barplots of enrichment tests (hypergeometric test) of features in a 
    sublist. The test is performed with stat_analysis.enrichment_test (see 
    its documentation for more details)

    Args:
        dict_cats (dict.): classification of each elements in a dictionary 
                of shape {category:[elements]}
                N.B.: the same element can be associated to multiple features
                Example: {"GOterm1":["GeneA","GeneB"],"GOterm2:["GeneA","GeneC"]}
        dict_features (dict.): feature corresponding to each element in a
                dictionary of shape {feature:[elements]}
                Example: {"act": ["GeneA","GeneC","GeneD"], "rep":["GeneB"]}
        targ_features (list or string): targeted feature(s)
        all_features (Optional [list or str.]): features that will be used to calculate
                the proportion, including targ_features
                (Default: all keys of the dict_features dictionary)
        targ_cats (Optional [list]): list of categories. The enrichment test is 
                performed for each catergory. (default: all keys of dict_cats)
        all_cats (Optional [list]): list of categories used to compute the global
                proportion and the expected number in the selection. 
                All_cats includes targ_cats. (default: all keys of dict_cats)
        min_nb_elements (Optional [int.]): Number of elements used as thresholds for the 
                feature selection. If there is stricly less than 
                min_nb_elements corresponding to a feature, the result 
                corresponding to this feature is not relevant and is 
                therefore neither returned nor reported in the output file.
                By default, min_nb_elements is set to 4.
        output_dir (Optional [str.]): output directory
        output_file (Optional [str.]): output filename for the proportion test data 
                (.txt) and the plot
        file_extension (Optional [str.]): Graphic file extension type (.pdf by default)
        xlabel (Optional [str.]): label for the x-axis (default: empty)
        ylabel (Optional [str.]): label for the y-axis (default: "Proportion")
        title (Optional [str.]): general title for the figure
        annot_star (Optional [bool.]): if True, the barplot will be annotated according
                to the p-values using stars annotaion (default: True)
        ymin (Optional [float]): y-axis bottom limit
        ymax (Optional [float]): y-axis top limit
        legend_text (Optional [str.]): Legend text (default: "Global proportion")
                If set to None, no legend will be plotted.
        legend_loc (Optional [str.]): Location of the legend such as 'upper right',
                'lower right', 'lower left', 'lower left' and 'best' 
                (default: 'best'). See matplotlib.pyplot.legend for more 
                options
        figsize (Optional [(float,float)]): width and height in inches (by default: (w,2.2)
                with w dependent on the number of categories)
        xticks_rotation (Optional [int.]): x-ticks labels rotation in degrees
        xticks_labels (Optional [list.]): x-ticks labels (by default: targ_cats)
        err_capsize (Optional [float]): Length of the error bar caps in points
        bar_linewidth (Optional [float]): Width of the bars edge. (default: 1.5)
        bar_width (Optional [float]): Width of the bars. (default dependent on the 
                number of categories. If less than 5 cats: 0.7)

    Example:
        >>> from GRATIOSA import plot_stat_analysis
        >>> dataX = {"GOterm1":["A","B","D","E","F"],
        ...           "GOterm2":["C","E"],
        ...           "GOterm3":["A","B","F","G","H","I","M"],
        ...           "GOterm4":["C","F","G","J"]}
        >>>  = ["A","E","I","F","G","H","J"]
        >>> plot_stat_analysis.plot_enrichment_test(
        ...                         dict_cats,dict_features,
        ...                         targ_feature=["act","None"],
        ...                         all_features=["act","None","rep","NA"],
        ...                         targ_cats=["GOterm1","GOterm2","GOterm3"],
        ...                         min_nb_elements=3,output_file="test1")
    '''

    df_res = stat_analysis.enrichment_test(dict_cats=dict_cats,
                                           dict_features=dict_features,
                                           targ_features=targ_features,
                                           all_features=all_features,
                                           targ_cats=targ_cats,
                                           all_cats=all_cats,
                                           min_nb_elements=min_nb_elements,
                                           output_dir=output_dir,
                                           output_file=output_file)
    print(df_res[["Category",
                  "Selected_gene_nb",
                  "Expected_selected_nb",
                  "Adj p-value (FDR)"]])
    print(df_res)
    res = df_res.set_index("Category").to_dict('index')

    prop = []
    ci = np.zeros((2, len(targ_cats)))
    for n, t in enumerate(targ_cats):
        prop.append(res[t]['Proportion'])
        ci[0, n] = res[t]['Proportion'] - res[t]["Prop_conf_int"][0]
        ci[1, n] = res[t]["Prop_conf_int"][1] - res[t]['Proportion']

    if len(targ_cats) < 5:
        wb = 0.7
    elif len(targ_cats) < 15:
        wb = 0.5
    else:
        wb = 0.4
    figsize = kwargs.get("figsize", (len(targ_cats) * wb/1.5+0.2, 2.2))
    plt.figure(figsize=figsize)

    bar_width = kwargs.get("bar_width", wb)
    bar_linewidth = kwargs.get("bar_linewidth", 1.5)
    err_capsize = kwargs.get("err_capsize", bar_width*5)
    plt.bar(targ_cats, prop, yerr=ci,
            color='white', edgecolor='black', ecolor='black',
            width=bar_width, linewidth=bar_linewidth, capsize=err_capsize)
    
    # gets the ylim to defines the necessary difference between the
    # errorbar and the annotation in order to be readable
    ax_y0, ax_y1 = plt.gca().get_ylim()
    dh = (ax_y1 - ax_y0) / 40

    max_ci = 0
    for n, t in enumerate(targ_cats):
        max_ci = max(max_ci, ci[1, n] + prop[n] + 4 * dh)
        if res[t]["Adj p-value (FDR)"] < 0.05 and annot_star:
            s = significance(res[t]["Adj p-value (FDR)"])
            plt.text(n, ci[1, n] + prop[n] + dh /
                     2, s, ha='center', fontsize=14)

    ymin = kwargs.get('ymin', plt.ylim()[0])
    ymax = kwargs.get('ymax', max_ci*1.1)
    plt.ylim(ymin, ymax)
    plt.ylabel(ylabel)
    if xlabel != "":
        plt.xlabel(xlabel)
    xticks_rotation = kwargs.get("xticks_rotation", 0)
    xticks_labels = kwargs.get("xticks_labels", targ_cats)
    plt.xticks(ticks=np.arange(len(targ_cats)),
               labels=xticks_labels,
               rotation=xticks_rotation)

    # plots an horizontal line corresponing to the Global_proportion
    plt.axhline(df_res["Global_proportion"][0], c="blue", ls='--',
                label=legend_text)
    if legend_text not in [None,"None"] :
        plt.legend(loc=legend_loc)
    plt.title(title)
    plt.tight_layout()
    plt.savefig(
        f"{output_dir}{output_file}{file_extension}",
        transparent=False,
        facecolor='white')
    plt.show()
    plt.close()


def plot_student_test(dict_data, cats="all",
                      alt_hyp="one-sided",
                      output_dir=f"{resdir}student_test/",
                      output_file=f"student_test{datetime.now()}",
                      file_extension=".pdf",
                      xlabel="",
                      ylabel="Mean(data)",
                      title="",
                      annot_brackets=True,
                      *args, **kwargs):
    '''
    Barplots of the student test computed with 
    stat_analysis.quantitative_data_student_test (see its documentation for more details)

    Args:
        dict_data (dict.): feature corresponding to each element in a
                dictionary of shape {category:list of datapoints}
        cats (Optional [list]): list of categories to compare
                (default: all keys of dict_data)
        alt_hyp (Optional ["two-sided" or "one-sided"]): alternative hypothesis.
                If "one-sided" is chosen, both one-sided tests will be 
                performed with the scipy.stats.ttest_ind function and the 
                smaller p-value will be kept. See scipy documentation for 
                more details. (default: "one-sided")
        output_dir (Optional [str.]): output directory
        output_file (Optional [str.]): output filename for the student test data (.txt)
                and the plot
        file_extension (Optional [str.]): Graphic file extension type (.pdf by default)
        xlabel (Optional [str.]): label for the x-axis (default: empty)
        ylabel (Optional [str.]): label for the y-axis (default: "Mean(data)")
        title (Optional [str.]): general title for the figure
        annot_brackets (Optional [bool.]): if True, the barplot will be annotated 
                according to the p-values using stars annotaion
                (default: True)
        brackets_linewidth (Optional [float.]): Linewidth of the brackets.(default: 1.5)
        ymin (Optional [float]): y-axis bottom limit
        ymax (Optional [float]): y-axis top limit
        figsize (Optional [(float,float)]): width and height in inches (by default: (w,2.2)
                with w dependent on the number of categories)
        xticks_rotation (Optional [int.]): x-ticks labels rotation in degrees
        xticks_labels (Optional [list]): x-ticks labels (by default: cats)
        err_capsize (Optional [float]): Length of the error bar caps in points
        bar_linewidth (Optional [float]): Width of the bars edge. (default: 2.0)
        bar_width (Optional [float]): Width of the bars. (default: 0.6) (default 
                dependent on the number of categories. If less than 5 cats: 0.7)

    Example:
        >>> from GRATIOSA import plot_stat_analysis
        >>> dict_data = {'a':[1,2,5,6,19], 'b':[10,24,4,15]}
        >>> plot_stat_analysis.plot_student_test(dict_data)
    '''
    res = stat_analysis.quantitative_data_student_test(dict_data, cats=cats,
                                                       alt_hyp=alt_hyp,
                                                       output_dir=output_dir,
                                                       output_file=output_file)

    cats = res["categories"]
    means = res["means"]

    ci = np.zeros((2, len(cats)))
    for c in np.arange(len(cats)):
        ci[0, c] = means[c] - res["confidence intervals"][c][0]
        ci[1, c] = res["confidence intervals"][c][1] - means[c]

    if len(cats) < 5:
        wb = 0.7
    elif len(cats) < 15:
        wb = 0.5
    else:
        wb = 0.4
    figsize = kwargs.get("figsize", (len(cats) * wb/1.5+0.2, 2.2))
    plt.figure(figsize=figsize)

    bar_width = kwargs.get("bar_width", wb)
    bar_linewidth = kwargs.get("bar_linewidth", 1.5)
    err_capsize = kwargs.get("err_capsize", bar_width*5)
    plt.bar([str(c) for c in cats], means, yerr=ci,
            color='white', edgecolor='black', ecolor='black',
            width=bar_width, linewidth=bar_linewidth, capsize=err_capsize)
    error_max = np.amax(res['confidence intervals'], axis=1)

    if annot_brackets:
        brackets_linewidth = kwargs.get("brackets_linewidth", 1.5)
        barplot_annotate_brackets(
            cats,
            error_max,
            res["p-values"],
            linewidth=brackets_linewidth)

    plt.ylabel(ylabel)
    if xlabel != "":
        plt.xlabel(xlabel)
    ymin = kwargs.get('ymin', plt.ylim()[0])
    ymax = kwargs.get('ymax', plt.ylim()[1])
    xticks_rotation = kwargs.get("xticks_rotation", 0)
    xticks_labels = kwargs.get("xticks_labels", cats)
    plt.xticks(ticks=np.arange(len(cats)),
               labels=xticks_labels,
               rotation=xticks_rotation)
    plt.xticks(rotation=xticks_rotation)
    plt.ylim(ymin, ymax)
    plt.axhline(0, color="black")
    plt.title(title)
    plt.tight_layout()
    plt.savefig(
        f"{output_dir}{output_file}{file_extension}",
        transparent=False,
        facecolor='white')
    plt.show()
    plt.close()


import os
import sys

# basedir is the path to the folder which contains the database.
try:
    basedir = os.environ['GRATIOSA_DB_PATH']
except KeyError:
    print("Please set the environment variable GRATIOSA_DB_PATH")
    sys.exit()

# resdir is the default output folder.
resdir = basedir + 'resdir/'



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats
from datetime import datetime
from GRATIOSA.useful_functions_transcriptome import *
from GRATIOSA.globvar import *
from GRATIOSA.Genome import Genome

class Transcriptome:
    '''
    The Transcriptome class is the key class for loading expression data and
    differential expression data for analysis and comparison. Typical data are
    data obtained with high throughput methods such as RNASeq and processed with
    common tools such as DeSeq2.

    Each Transcriptome instance has to be initialized with an organism name
    
        >>> tr = Transcriptome.Transcriptome("dickeya")
    '''

    def __init__(self, name):
        self.name = name

    def load_expression(self):
        '''
        loads expression data (such as log2rpkm) from files that are in the
        /expression/ directory.

        Creates:
            self.genes[locus].expression (dict.): new attribute of Gene 
                    instances related to the Transcriptome instance given as 
                    argument. Dictionary of shape 
                    {condition: expression level (float.)}
            self.genes_valid_expr(dict.): Dictionary of shape 
                    {condition: list of genes having an expression value}

        Note:
            The data importation requires an expression.info file, containing
            column indices of each information in the data file and some 
            additional information, in the following order:
            [0] Condition [1] Filename [2] Locus_tag column
            [3] Expression column [4] is Log ? (boolean) [5] Separator

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an 
            annotation to your Transcriptome instance with the following commands 
            before using this method:
            >>> from GRATIOSA import Genome, Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> g = Genome.Genome(tr.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> tr.genes = g.genes

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("dickeya")
            >>> tr.load_expression()
            >>> tr.genes["Dda3937_00004"].expression
            {'WT_nov_stat(E10)_rpkm': 66.6413789332929,
            'WT_stat(E3)': 6.980245227157392,
            'WT_PGA_stat(E13)_rpkm': 13.9428053948966}
        '''
        if not hasattr(self, 'genes_valid_expr'):
            self.genes_valid_expr = {}
        if not hasattr(self, "genes"):
            gen = Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes

        path2dir = f"{basedir}data/{self.name}/expression/"
        if os.path.exists(path2dir + "expression.info"):
            with open(path2dir + "expression.info", "r") as f:
                skiphead = next(f)  # skip head
                for header in f:
                    header = header.strip().split('\t')
                    path2file = path2dir + header[1]
                    print(f"Loading condition: {header[0]}")
                    self.genes_valid_expr[header[0]] = add_expression_to_genes(
                        genes_dict=self.genes,
                        cond=header[0],
                        filename=path2file,
                        tag_col=int(header[2]),
                        expr_col=int(header[3]),
                        is_log=header[4],
                        separator=header[5])
            f.close()
        else:
            print("No expression.info file, please create one")

    def compute_fc_from_expr(self, ctrls, conds, condname):
        '''
        Compute_fc_from_expr computes, for each gene:
            * the mean of expression values (such as log2rpkm) in a list of 
              controls replicates (ctrls)
            * the mean of expression values in a list of tests replicates (conds)
            * the fold-change ie the difference between the mean control 
              and the mean test values
            * the pvalue of the Student-test for these two means

        Creates:
            * self.genes[locus].fc_pval[condname] (tuple)
                    new attribute of Gene instances related to the 
                    Transcriptome instance given as argument.Tuple of shape 
                    (fold-change,pvalue).
            * self.genes_valid_fc[condname] (list.)
                    list of genes having a fold-change and a pvalue corresponding 
                    to the new condition

        Args:
            conds (list of str.): list of expression conditions to use as test
            ctrls (list of str.): list of expression conditions to use as control
            condname (str.): condition name to use for the new fold-changes
                             and p-values
       
        Note:
            The t-test is computed with scipy.stats.ttest_ind

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an 
            annotation to your Transcriptome instance with the following commands 
            before using this method:
            >>> from GRATIOSA import Genome, Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> g = Genome.Genome(tr.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> tr.genes = g.genes

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> tr.compute_fc_from_expr(["Ctrl1","Ctrl2"],
                                        ["Test1","Test2"],
                                        ["FC_Test12_Ctrl12"])
            >>> tr.genes["b0002"].fc_pval["FC_Test12_Ctrl12"]
            (-1.786, 0.003125)
        '''
        if not hasattr(self, 'genes_valid'):
            self.genes_valid = {}
        self.load_expression()  # load expression values

        genes_val = []  # list containing all genes having valid expr
        for genename in self.genes.keys():
            gene = self.genes[genename]
            ctrlvals = []  # list of control values for that gene
            testvals = []  # list of test values for that gene
            try:
                for ctrl in ctrls:
                    ctrlvals.append(gene.expression[ctrl])
                for cond in conds:
                    testvals.append(gene.expression[cond])
                # add FC (meantest - meanctrl) and p-values (Student test)
                gene.add_fc_pval_cond(
                    np.mean(testvals) -
                    np.mean(ctrlvals),
                    condname,
                    stats.ttest_ind(
                        ctrlvals,
                        testvals,
                        equal_var=False)[1])
                genes_val.append(genename)
            except BaseException:
                pass

        self.genes_valid[condname] = genes_val

    def load_fc_pval(self):
        '''
        loads Fold-changes and p-values (if available) from data files that 
        are in the /fold_changes/ directory.

        Creates:
            * self.genes[locus].fc_pval (dict.): new attribute of Gene instances
                    related to the Transcriptome instance given as argument.
                    Dictionary of shape {condition: (fold-change,pvalue)}.
                    If no p-value is available in the data file, 0 is set as
                    p-value.
            * self.genes_valid_fc (dict.): Dictionary of shape 
                    {condition: list of genes having at least a fold-change}

        Note:
            The data importation requires a fc.info file, containing
            column indices of each information in the data file and some 
            additional information, in the following order:
            [0] Condition [1] Filename [2] Locus_tag column
            [3] FC column [4] Separator [5] Startline [6] P-value column

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an 
            annotation to your Transcriptome instance with the following commands 
            before using this method:
            >>> from GRATIOSA import Genome, Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> g = Genome.Genome(tr.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> tr.genes = g.genes

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> tr.load_fc_pval()
            >>> tr.genes["b0002"].fc_pval
            {'WT': (-1.786, 0.003125),
             'osmotic': (-1.81603942323042, 0.0),
             'heat': (1.62203513077067, 0.0)}
        '''
        self.genes_valid_fc = {}

        if not hasattr(self, "genes"):
            gen = Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes

        path2dir = f"{basedir}data/{self.name}/fold_changes/"
        if os.path.exists(path2dir + "fc.info"):
            with open(path2dir + "fc.info", "r") as f:
                skiphead = next(f)  # skip head
                for header in f:
                    header = header.strip()
                    header = header.split('\t')
                    print(f"Loading condition: {header[0]}")
                    try:  # if p-value column specified works
                        self.genes_valid_fc[header[0]] = load_fc_pval_cond(
                            genes_dict=self.genes,
                            filename=path2dir + header[1],
                            condition=header[0],
                            tag_col=int(header[2]),
                            fc_col=int(header[3]),
                            separator=header[4],
                            start_line=int(header[5]),
                            p_value=int(header[6]))
                    except BaseException:  # otherwise, set pvalue = 0
                        self.genes_valid_fc[header[0]] = load_fc_pval_cond(
                            genes_dict=self.genes,
                            filename=path2dir + header[1],
                            condition=header[0],
                            tag_col=int(header[2]),
                            fc_col=int(header[3]),
                            separator=header[4],
                            start_line=int(header[5]))
            f.close()
        else:
            print("No fc.info file, please create one")

    def compute_state_from_fc(self, thresh_pval=0.05, thresh_fc=0):
        '''
        Loads Fold-changes and p-values data and computes genes state from 
        these data. If no p-values is available in the data file, the default 
        value of 0 is assigned to each gene's pvalue.

        A gene is considered:
            * `activated` if its FC is above the FC threshold given as argument
              and its pvalue is below the pvalue threshold given as 
              argument ie. FC > thresh_FC and pval < thresh_pval
            * `repressed` if its FC is below the opposite of the FC threshold
              and its pvalue is below the pvalue threshold
              ie. FC < -thresh_FC and pval < thresh_pval
            * `not affected` either if its pvalue is above the threshold,
              or if its FC is between the - thresh_FC and + thresh_FC

        Creates:
           
            * self.statesFC (dict. of dict.)
                    New attribute of the  Transcriptome instance. 
                    Dictionary containing one subdictionary per condition 
                    listed in fc.info. Each subdictionary contains the list of 
                    genes corresponding to  each state ('act', 'rep', 'non' or 
                    'null').
            * self.genes[locus].fc_pval (dict.) 
                    new attribute of Gene instances
                    related to the Transcriptome instance given as argument.
                    Dictionary of shape {condition: (FC,pvalue)}.
            * self.genes[locus].state (dict.) 
                    new attribute of Gene instances
                    related to the Transcriptome instance given as argument.
                    Dictionary of shape {condition: state} with state either
                    * `act` if the gene is activated
                    * `rep` if the gene is repressed
                    * `non` if the gene is not affected
                    * `null` if the gene is not present in the data


        Args:
            thresh_pval (Float): pvalue threshold used for the genes 
                    classification
            thresh_fc (Float): fold-change threshold used for the genes 
                    classification
    
        Note:
            The FC and pvalues data importation requires an FC.info file, in the
            fold_changes directory, containing column indices of each information 
            in the data file, in the following order:
            [0] Condition [1] Filename [2] Locus_tag column [3] FC columns
            [4] Separator [5] File start line [6] P-value column
            See load_fc_pval for more details about the data importation.

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an 
            annotation to your Transcriptome instance with the following commands 
            before using this method:
            >>> from GRATIOSA import Transcriptome, Genome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> g = Genome.Genome(tr.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> tr.genes = g.genes

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> tr.compute_state_from_fc()
            >>> tr.genes['b0001'].state
            {'osmotic': 'rep', 'acidic_1mn': 'act'}
            >>> tr.genes['b0001'].fc_pval
            {'osmotic': (-1.73717046009437, 0.0), 'acidic_1mn': (1.73, 0.0)}
            >>> tr.statesFC['osmotic']['rep']
            ['b0001', 'b0002', 'b0003', 'b0004',...]
        '''

        if not hasattr(self, 'genes_valid_fc'):
            self.load_fc_pval()
        if not hasattr(self, "genes"):  # if no genes loaded
            # try to load them
            gen = Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes
        if not hasattr(self, "statesFC"):
            self.statesFC = {}

        for cond_fc in self.genes_valid_fc.keys():
            self.statesFC[cond_fc] = {"rep": [], "act": [], "non": []}
        for gene in self.genes.keys():
            g = self.genes[gene]
            for cond_fc in self.genes_valid_fc.keys():
                try:
                    if g.fc_pval[cond_fc][1] <= thresh_pval:
                        if g.fc_pval[cond_fc][0] < 0 - thresh_fc:
                            g.add_state_cond(cond_fc, 'rep')
                            self.statesFC[cond_fc]['rep'].append(gene)
                        elif g.fc_pval[cond_fc][0] > 0 + thresh_fc:
                            g.add_state_cond(cond_fc, 'act')
                            self.statesFC[cond_fc]['act'].append(gene)
                        else:
                            g.add_state_cond(cond_fc, 'non')
                            self.statesFC[cond_fc]['non'].append(gene)
                    else:
                        g.add_state_cond(cond_fc, 'non')
                        self.statesFC[cond_fc]['non'].append(gene)
                except BaseException:
                    if not hasattr(self.statesFC[cond_fc], 'null'):
                        self.statesFC[cond_fc]['null'] = [gene]
                    else:
                        self.statesFC[cond_fc]['null'].append(gene)
                    g.add_state_cond(cond_fc, 'null')

    def load_rnaseq_cov(self, cond="all", compute_from_bam=False):
        '''
        load_rnaseq_cov loads a RNASeq coverage to a Transcriptome instance
        either from:
        * .npz files (described in a cov.info file and computed when new 
          data are loaded for the first time)  which are in the /rnaseq_cov/ 
          directory,
        * coverage files which are in the /rnaseq_cov/ directory and are 
          described in cov_txt.info file,
        * paired-end .bam reads files which are in the /rnaseq_reads/ 
          directory  and are treated with 
          useful_functions_transcriptome.cov_from_reads function.

        Creates 2 new attributes of the Transcriptome instance:
            * self.rnaseq_cov_pos: 
                    dictionary of shape {condition: cov+}
                    with cov+ an array containing one signal data per genomic
                    position for the + strand (forward coverage)
            * self.rnaseq_cov_neg: 
                    idem for the - strand (reverse coverage)

        

        Args:
            cond (Optional [list of str.]): selection of one or several 
                    conditions (1 condition corresponds to 1 data file).
                    By default: cond ='all' ie all available coverages are 
                    loaded.
            compute_from_bam (Boolean.): if True, computes, using 
                    useful_functions_transcriptome.cov_from_reads, coverage 
                    from paired-end .bam reads files that are, with a 
                    bam_files.info file, in the /rnaseq_reads/ directory.
                    WARNING: Works only with paired-end files

        Note:
            To use directly new coverages data, both forward and reverse coverage 
            data files are required. These files contain the coverage for each 
            genomic position (one line = one genomic position, no position can be 
            ommited), therefore if you use bedtools genomecov, please use the -d 
            option.
            The importation of these new data requires a cov_txt.info 
            file, in the /rnaseq_cov/ directory, containing the following 
            information:
            [0] Condition
            [1] Reverse coverage filename
            [2] Forward coverage filename
            [3] Startline (has to be the same for both coverage files)
            [4] Column containing the coverage data

            Coverages from the reverse file will be added to the Transcriptome
            instance as "rnaseq_cov_neg" attribute and coverages from the forward
            file will be the "rnaseq_cov_pos"attribute.

        Note:
            To compute coverages data from reads files, the data files and a
            bam_files.info file have to be in the /rnaseq_reads/ directory.
            The bam_files.info files contains:
            [0] Condition [1] Reads filename

        Note:
            During the first importation of new coverages data, a .npy file
            containing the data is created and will enable a faster data loading
            for the next data importation. A cov.info file containing information
            for this fast importation is also in the /rnaseq_cov/ directory and
            contains in the following columns:
            [0] Condition [1] Coverage filename
            This cov.info file automatically completes itself when the data are
            loaded for the first time.

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            # To load only the coverage for one condition named "WT"
            # in the cov.info or cov_txt.info file
            >>> tr.load_rnaseq_cov(["WT"]) 
            >>> tr.rnaseq_cov_neg["WT"]
            array([0., 0., 0., ..., 0., 0., 0.])
            >>> tr.rnaseq_cov_pos["WT"]
            array([0., 0., 0., ..., 0., 0., 0.])
            # To compute the coverage from a new .bam files obtained from 
            # paired-end data for example with: tophat2 genome/MG1655 
            # Test_1.fastq.gz Test_2.fastq.gz
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> tr.load_rnaseq_cov(compute_from_bam=True)
            >>> tr.rnaseq_cov_pos["Test"]
            array([10., 10., 10., ..., 0., 0., 0.])
            >>> tr.rnaseq_cov_pos["Test"]
            array([0., 0., 0., ..., 20., 20., 20.])
        '''
        if not hasattr(self, "rnaseq_cov_pos"):
            self.rnaseq_cov_pos = {}  # cov on + strand
            self.rnaseq_cov_neg = {}  # cov on - strand

        # conversion from str to list (useful if only 1 condition is selected)
        if isinstance(cond, str):
            cond = [cond]

        path2dir = f"{basedir}data/{self.name}/rnaseq_cov/"

        if compute_from_bam:
            process_bam_paired_end(self)
            cov_from_reads(self)

        if os.path.exists(f"{path2dir}cov_txt.info"):
            if not os.path.exists(f"{path2dir}cov.info"):
                file = open(f"{path2dir}cov.info", 'w')
                file.write('Condition\tCov file\n')
            else:
                file = open(f"{path2dir}cov.info", 'a')
            with open(f"{path2dir}cov_txt.info", "r") as f:
                header = next(f)
                for line in f:
                    line = line.strip('\n').split('\t')
                    print(f'Updating cov.info with the new data: {line[0]}')
                    # create .npy from .txt
                    rnaseq_cov_neg = np.loadtxt(
                        path2dir + line[1], usecols=[int(line[4])], skiprows=int(line[3]))
                    rnaseq_cov_pos = np.loadtxt(
                        path2dir + line[2], usecols=[int(line[4])], skiprows=int(line[3]))
                    # save .npy into .npz
                    np.savez(
                        f"{path2dir}{line[0]}_cov.npz",
                        cov_pos=rnaseq_cov_pos,
                        cov_neg=rnaseq_cov_neg)
                    # update cov.info
                    file.write(f"{line[0]}\t{line[0]}_cov.npz\n")
            with open(f"{path2dir}cov_txt.info", "w") as f:
                f.write(header)
            f.close()
            file.close()

        if os.path.exists(f"{path2dir}cov.info"):
            with open(f"{path2dir}cov.info", "r") as f:
                header = next(f)
                loaded_cond = []
                for line in f:
                    line = line.strip().split('\t')
                    # For each selected condition:
                    if cond == ["all"] or line[0] in cond:
                        print('Loading condition', line[0])
                        loaded_cond.append(line[0])
                        # load attributes
                        self.rnaseq_cov_neg[line[0]] = np.load(
                            path2dir + line[1])["cov_neg"]
                        self.rnaseq_cov_pos[line[0]] = np.load(
                            path2dir + line[1])["cov_pos"]
                if cond != ["all"]:
                    unloaded = set(loaded_cond) ^ set(cond)
                    if len(unloaded) != 0:
                        print(f"{unloaded} data were not found, please check .info files")
            f.close()

        else:
            print(
                'cov.info not available nor cov_txt.info, please check /rnaseq_cov/ folder')

    def load_cov_start_end(self, compute_from_bam=False):
        '''
        Useful to locate TSS and TTS, load_cov_start_end loads the density of
        RNA fragment starts and ends from .npz files described in
        cov_start_stop.info, in the /rnaseq_cov/ directory.
        .npz files and .info files are obtained using the
        useful_functions_transcriptome.cov_start_stop_from_reads

        Creates 2 new attributes of the Transcriptome instance:
            * self.cov_start (Dict. of dict.)
                    dictionary containing one subdictionary per condition.
                    Each subdictionary has 2 keys:
                    "0" for the start positions on the - strand and
                    "1" for the start positions on the + strand
                    and the subdictionary values are arrays containing
                    all start positions on the corresponding strand
            * self.cov_end (Dict. of dict.)
                    idem for the end positions

        Args:
            compute_from_bam (Boolean): if True, computes, using 
                    useful_functions_transcriptome.cov_start_stop_from_reads,
                    coverage from paired-end .bam reads files that are, with 
                    a bam_files.info file, in the /rnaseq_reads/ directory.

        Note:
            To compute new coverages data from paired-end .bam reads files, the 
            data files and a bam_files.info file have to be in the /rnaseq_reads/ 
            directory and the input argument "compute_from_bam" has to be set to 
            "True".The bam_files.info files contains:
            [0] Condition [1] Reads filename
            During the importation of these new data, a corresponding .npz files
            is created and the cov_start_stop.info file automatically completes 
            itself. For the next importations of these data, only the npz. and 
            the cov_start_stop.info will be loaded, thus allowing faster loading.

        Warning:
            compute_from_bam works only with paired-end files

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> tr.load_cov_start_end(compute_from_bam=True)
            >>> tr.cov_end["WT_paired"]
            {0: array([0, 0, 0, ..., 0, 0, 0]),
             1: array([0, 0, 0, ..., 0, 0, 0])}
        '''

        if compute_from_bam:
            process_bam_paired_end(self)
            cov_start_stop_from_reads(self)

        self.cov_start = {}
        self.cov_end = {}
        path2dir = f"{basedir}data/{self.name}/rnaseq_cov/"
        if os.path.exists(f"{path2dir}cov_start_stop.info"):
            with open(f"{path2dir}cov_start_stop.info", "r") as f:
                header = next(f)
                for line in f:  # for each condition
                    line = line.strip().split('\t')
                    print('Loading condition', line[0])
                    # load attributes
                    self.cov_start[line[0]] = {}
                    self.cov_end[line[0]] = {}
                    self.cov_start[line[0]][0] = np.load(
                        path2dir + line[1])["cov_start_neg"]
                    self.cov_start[line[0]][1] = np.load(
                        path2dir + line[1])["cov_start_pos"]
                    self.cov_end[line[0]][0] = np.load(
                        path2dir + line[1])["cov_end_neg"]
                    self.cov_end[line[0]][1] = np.load(
                        path2dir + line[1])["cov_end_pos"]
            f.close()

        else:
            print('cov_start_stop.info not available please check /rnaseq_cov/ folder')

    def compute_log2rpkm_from_cov(self, cond='all', before=100):
        '''
        Compute_log2rpkm_from_cov method:
            * loads the RNASeq coverage (using load_rnaseq_cov method)
            * computes and loads RPKM value on the Gene instances
              (using Gene.add_single_rpkm )
            * saves the log2(RPKM) data in a new .csv file
              (1 file for all conditions, with 1 column per condition)
            * adds the new .csv file informations in the expression.info file 
              in the /expression/ directory

        Creates:
            * self.genes[locus].rpkm (dict.):
                    new attribute of Gene instances related to the 
                    Transcriptome instance given as argument.
                    Dictionary of shape {condition: RPKM value}, containing 
                    the RPKM for each computed condition.

            * log2rpkm_from_cov_{datetime.now()()}.csv: csv file containing the
                    log2(RPKM) values (1 row per gene, 1 column per condition).
                    If a RPKM is equal to 0, log2RPKM is set to 1.
                    This file is saved in the /expression/ directory and is
                    automatically added in the expression.info file.

        Args: 
            cond (list of str.):
                    selection of one or several RNASeq coverage condition
                    By default: cond ='all' ie all available coverages are
                    converted. All selected conditions have to be listed in
                    cov.info file. If one of the chosen condition is already
                    registered in the expression.info file in the expression/
                    directory, this condition is ignored.
            before (int.):
                    number of bps before the gene start to take into account

        Warning: 
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an 
            annotation to your Transcriptome instance with the following commands 
            before using this method:
            >>> from GRATIOSA import Genome, Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> g = Genome.Genome(tr.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> .tr.genes = g.genes

        Example:
            >>> from GRATIOSA import Transcriptome
            >>> tr = Transcriptome.Transcriptome("ecoli")
            >>> tr.compute_log2rpkm_from_cov(cond="WT")
            >>> tr.genes["b0002"].rpkm["WT"]
            142
        '''

        # Loading the annotation and the RNASeq coverage
        if not hasattr(self, "genes"):
            gen = Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes
        if not hasattr(self, "rnaseq_cov_pos"):
            self.load_rnaseq_cov()

        # Checking if each condition is in cov.info and not in expression.info
        path2dir = f"{basedir}data/{self.name}/expression/"
        if isinstance(cond, str):
            cond = [cond]
        if cond == ["all"]:
            cond = self.rnaseq_cov_pos.keys()
        else:
            condpb = set(cond) - set(self.rnaseq_cov_pos.keys())
        if condpb:
            print(f"Following conditions are not in cov.info: {condpb}")

        if os.path.exists(f"{path2dir}expression.info"):
            existing_cond = []
            with open(f"{path2dir}expression.info", 'r')as f:
                header = next(f)
                for line in f:
                    line = line.strip('\n').split('\t')
                    existing_cond.append(line[0])
            f.close()

        cond2convert = list(set(cond) - set(existing_cond) - condpb)
        ignoredcond = set(cond) & set(existing_cond)
        if ignoredcond:
            print(f"Following conditions are already in expression.info: {ignoredcond})")

        # Computing the RPKM
        if cond2convert:
            for g in self.genes.keys():  # for each gene
                try:
                    if self.genes[g].strand:
                        # gene in + strand
                        for c in cond2convert:  # for each condition of cov
                            self.genes[g].add_single_rpkm(c, np.mean(self.rnaseq_cov_pos[c][(
                                self.genes[g].left - before):self.genes[g].right]), np.sum(self.rnaseq_cov_pos[c]) + np.sum(self.rnaseq_cov_neg[c]))
                    elif not self.genes[g].strand:
                        # gene in - strand
                        for c in cond2convert:
                            self.genes[g].add_single_rpkm(c, np.mean(self.rnaseq_cov_neg[c][self.genes[g].left:(
                                self.genes[g].right + before)]), np.sum(self.rnaseq_cov_pos[c]) + np.sum(self.rnaseq_cov_neg[c]))
                except Exception as e:
                    print(f'WARNING: {g} , {cond}, {e}')
                    pass

        # Saving log2RPKM in a .csv file and completing the expression.info
        # file
            file = []
            for g in self.genes.keys():
                g_log2rpkm = []
                for c in cond2convert:
                    if self.genes[g].rpkm[c] != 0:
                        g_log2rpkm.append(np.log2(self.genes[g].rpkm[c]))
                    else:
                        g_log2rpkm.append(1)
                file.append(g_log2rpkm)
            df = pd.DataFrame(data=file, columns=cond2convert,
                              index=self.genes.keys())
            date = str(datetime.now())[:-10].replace(" ", "_")
            filename = f"log2rpkm_from_cov_{date}.csv"
            df.to_csv(path2dir + filename, sep=',', index=True)

            if not os.path.exists(f"{path2dir}expression.info"):
                f = open(f"{path2dir}expression.info", 'w')
                f.write('Condition\tFilename\tTagColumn\tExprColumn\tIslog(yes/no)\tSeparator\n')
            else:
                f = open(f"{path2dir}expression.info", 'a')

            for n, c in enumerate(cond2convert):
                f.write(f'{c}\t{filename}\t0\t{n+1}\tyes\t,\n')
        else:
            print("All selected conditions are already in expression.info")


#! /usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import numpy as np
import pandas as pd
from pathlib import Path
from GRATIOSA.globvar import *
from GRATIOSA.useful_functions_HiC import *
from GRATIOSA import Genome

class HiC:
    '''
    The HiC class is used to gather spatial information obtained from HiC data 
    analysis with tools such as Chromosight. The two patterns analyzed here are 
    borders (between two topological domains) and loops (more specifically the 
    positions of the two regions that come into contact).

    Each HiC instance has to be initialized with an organism name
    
    Example:
        >>> from GRATIOSA import HiC
        >>> hc = HiC.HiC("dickeya")
    '''

    def __init__(self, name):
        self.name = name

    def load_hic_borders(self, cond="all"):
        """
        load_hic_borders imports a list of borders from a data file (typically
        a borders.tsv data file obtained with Chromosight) containing at least,
        for each border, its position (bin number) and optionally its identification 
        score (pearson correlation coefficient between the border kernel and the 
        detected pattern), pvalue and qvalue.
        Creates 2 new attributes of the HiC instance:
        * self.borders (dict. of dict.) 
                dictionary containing one subdictionary per condition 
                with the shape self.borders[cond]={bin_nb: {"score": 
                score, "pval": pvalue, "qval": qvalue,"binsize":binsize}}.
                N.B.: If data are binned at 2kb, the bin with number 10 
                corresponds to data between 20000 and 22000.    
        * self.borders_pos (dict. of dict)
                dictionary containing one subdictionary per 
                condition, with the shape self.borders_pos[cond] =
                {'borders':list of positions that are in a border
                'no_borders':list of positions that are not in a border}

        Args:
            cond (Optional [list of str.]): selection of one or several
                    conditions (1 condition corresponds to 1 data file and 
                    each condition has to be listed in the border.info file).
                    By default: cond ='all' ie all available conditions are 
                    loaded.
        Note:
            The data importation requires a borders.info file that contains the
            column indices of each information in the data file and some
            additional information, in the following order:
            * (required) [0] Condition, [1] Filename, [2] Startline,
              [3] Separator, [4] Bin, [5] Binsize (in b)
            * (optional) [6] Score, [7] Pvalue,[8] Qvalue

        Note:    
            borders.info and data file have to be in the /HiC/Borders/ directory

        Note: 
            See Chromosight documentation for more details about the input format.

        Warning:
            Make sure that the version of the genome is the same as
            the version used in the HiC analysis workflow

        Example:
            >>> from GRATIOSA import HiC
            >>> hc = HiC.HiC("dickeya")
            >>> hc.load_hic_borders("WT")
            >>> hc.borders_pos['WT']['no_borders'][0:5]
            [2001,2002,2003,2004,2005]
        """

        # gets the path to data and .info files
        path2dir = f"{basedir}data/{self.name}/HiC/Borders/"

        if isinstance(cond, str):
            cond = [cond]

        # tries to open borders.info file
        if Path(f"{path2dir}borders.info").exists():
            with open(f"{path2dir}borders.info", "r") as f:
                skiphead = next(f)
                # 1 line in borders.info corresponds to 1 condition (1 data
                # file)
                for line in f:
                    line = line.strip('\n').split('\t')
                    # loads data file information for this condition
                    if cond == ["all"] or line[0] in cond:

                        path2file = f"{basedir}data/{self.name}/HiC/Borders/{line[1]}"
                        startline, sep = int(line[2]), line[3]
                        bin_col = line[4]

                        # loads binsize
                        binsize = int(line[5])

                        # loads column indices of each optional information
                        print(line[0])
                        score_col, pval_col, qval_col = None, None, None
                        if len(line) > 6:
                            try : 
                                score_col = int(line[6])
                            except :
                                pass
                        if len(line) > 7:
                            try : 
                                pval_col = int(line[7])
                            except :
                                pass
                        if len(line) > 8:
                            try : 
                                qval_col = int(line[8])
                            except :
                                pass

                        # creates the new attribute using the load_HiC_site_cond function
                        # from useful_functions_HiC
                        if not hasattr(self, "borders"):
                            self.borders = {}  # creates the new attribute
                        self.borders[line[0]] = load_HiC_site_cond(
                            'borders', path2file, startline, sep, bin_col, binsize,
                            score_col=score_col, pval_col=pval_col, qval_col=qval_col)
        else:
            print("No borders.info, unable to load borders")

        if cond == ['all']:
            cond = list(self.borders.keys())

        if not hasattr(self, "length"):
            g = Genome.Genome(self.name)
            g.load_seq()
            self.length = g.length

        if not hasattr(self, "borders_pos"):
            self.borders_pos = {}

        for name in cond:
            print(f"Loading borders in: {name}")
            borders_list = list(self.borders[name].keys())
            binsize = self.borders[name][borders_list[0]]["binsize"]
            pos_borders = []
            is_border = [False] * self.length
            for border in borders_list:
                ps = np.arange(border, border + binsize + 1)
                pos_borders.extend(ps)
                for p in ps:
                    is_border[p] = True
            pos_no_borders = set(np.arange(self.length)) - set(pos_borders)
            self.borders_pos[name] = {
                "borders": list(pos_borders),
                "no_borders": list(pos_no_borders)}

    def load_hic_loops(self, cond="all"):
        """
        load_hic_loops imports a list of loops from a data file (typically
        a loops.tsv data file obtained with Chromosight) containing, for each 
        loop, at least the loop coordinates (2 bin numbers) and optionally the 
        loop score (pearson correlation coefficient between the loop kernel and 
        the detected pattern), the p-value and the q-value.

        Creates 2 new attributes of the HiC instance:
            * self.loops (dict. of dict.)
                dictionary containing one subdictionary per condition with 
                the shape self.loops[cond] =
                {(bin1_nb,bin2_nb): {"score": score, "pval": pvalue,
                "qval": qvalue, "binsize":binsize}}
                N.B.: If data are binned at 2kb, the bin with number 10 
                corresponds to data between 20000 and 22000.
            * self.loops_pos (dict. of dict)
                dictionary containing one subdictionary per condition, with 
                the shape self.loops_pos[cond] =
                {'loops':list of positions that are in a loop,
                'no_loops':list of positions that are not in a loop}

        Args:
            self (HiC instance)
            cond (Optional [list of str.]): selection of one or several 
                    conditions (1 condition corresponds to 1 data file and 
                    has to be in the loops.info file). By default: 
                    cond ='all' ie all available data are loaded.

        Note:
            The data importation requires a loops.info file that contains the
            column indices of each information in the data file and some 
            additional information, in the following
            order:
            * (required) [0] Condition, [1] Filename, [2] Startline,
                  [3] Separator, [4] Bin1, [5] Bin2, [6] Binsize (in b),
            * (optional) [7] Score, [8] Pvalue,[9] Qvalue
        
        Note: 
            loops.info and data file have to be in the /HiC/Loops/ directory
        
        Note: 
            See Chromosight documentation for more details about the input format.

        Warning:
            Make sure that the version of the genome is the same as
            the version used in the HiC analysis workflow

        Example:
            >>> from GRATIOSA import HiC
            >>> hc = HiC.HiC("dickeya")
            >>> hc.load_hic_loops()
            >>> hc.loops_pos['WT']['no_borders'][0:5]
            [2001,2002,2003,2004,2005]
            >>> hc.loops['WT1']
            {(42000, 80000): {'binsize': 2000},
             (238000, 252000): {'binsize': 2000}
             ...}
        """
        # gets the path to data and .info files
        path2dir = f"{basedir}data/{self.name}/HiC/Loops/"

        if isinstance(cond, str):
            cond = [cond]

        # tries to open loops.info file
        if Path(f"{path2dir}loops.info").exists():
            with open(f"{path2dir}loops.info", "r") as f:
                skiphead = next(f)  # skip head

                # 1 line in loops.info corresponds to 1 condition (1 file)
                for line in f:
                    line = line.strip('\n')
                    line = line.split('\t')

                    if cond == ["all"] or line[0] in cond:

                        # loads data file information for this condition
                        path2file = f"{path2dir}{line[1]}"
                        startline, sep = int(line[2]), line[3]
                        bin1_col, bin2_col = line[4], line[5]

                        # loads binsize
                        binsize = int(line[6])

                        # loads column indices of each optional information
                        score_col, pval_col, qval_col = None, None, None
                        if len(line) > 7:
                            try : 
                                score_col = int(line[7])
                            except :
                                pass
                        if len(line) > 8:
                            try : 
                                pval_col = int(line[8])
                            except :
                                pass
                        if len(line) > 9:
                            try : 
                                qval_col = int(line[9])
                            except :
                                pass
                        # creates the new attribute using the load_HiC_site_cond 
                        # function from useful_functions_HiC

                        if not hasattr(self, "loops"):
                            self.loops = {}  # creates the new attribute
                        self.loops[line[0]] = load_HiC_site_cond(
                            "loops", path2file, startline, sep, bin1_col, binsize,
                            bin2_col=bin2_col, score_col=score_col,
                            pval_col=pval_col, qval_col=qval_col)

        else:
            print("No loops.info, unable to load loops")
        if cond == ['all']:
            cond = list(self.loops.keys())

        if not hasattr(self, "loops_pos"):
            self.loops_pos = {}

        if not hasattr(self, "length"):
            g = Genome.Genome(self.name)
            g.load_seq()
            self.length = g.length

        # register loops genomic positions in the loops_pos attribute
        for c in cond:
            print(f"Loading loops in: {c}")

            self.loops_pos[c] = {"loops": [], "no_loops": []}
            loops_list = list(self.loops[c].keys())
            binsize = self.loops[c][loops_list[0]]["binsize"]

            is_loop = [False] * self.length

            for loop in loops_list:
                for pos in list(np.arange(loop[0], loop[0] + binsize + 1)) + list(np.arange(loop[1], loop[1] + binsize + 1)):
                    is_loop[pos] = True

            for x in np.arange(self.length):
                if is_loop[x]:
                    self.loops_pos[c]["loops"].append(x)
                else:
                    self.loops_pos[c]["no_loops"].append(x)

    def load_loops_genes(self, cond='all', window=0):
        """
        load_loops_genes imports a list of loops determined with HiC from a 
        data file (typically a loops.tsv data file obtained with Chromosight)
        containing at least the loops coordinates (2 bin numbers) and 
        determines the list of genes overlapping one of the loops positions.

        Creates:
            self.loops_pos (dict. of dict.) 
                    new attribute of the Genome 
                    instance. Dictionary containing one dictionary per 
                    condition listed in loops.info. Each subdictionary has 
                    2 keys: "loops" and "no_loops" and contains the 
                    corresponding lists of genomic positions.
            self.loops_genes (dict. of dict.) 
                    New attribute of the Genome instance. 
                    Dictionary containing one dictionary per condition listed 
                    in loops.info. The key of this dictionary contains the 
                    condition name and the chosen window size (for example 
                    "cond_w100b" for a 100b window).
                    Each subdictionary has 2 keys: "loops" and "no_loops" and
                    contains the corresponding lists of genes.
                    For example, self.loops_genes[cond_w0b]["loops"] returns
                    the list of genes that overlap the loops positions.
            self.genes[locus].is_loop (dict.) 
                    New attribute of Gene instances related to the Genome
                    instance given as argument.Dictionary of shape
                    {condition:boolean}. The boolean is True if the gene
                    overlaps the loop positions (window included).

        Args:
            cond (Optional [list of str.]): selection of one or several 
                    conditions (1 condition corresponds to 1 data file and 
                    has to be in the loops.info file). By default: 
                    cond ='all' ie all available data are loaded.
            per_genes(Optional [Bool.]): if True, determines the list of 
                    genes overlapping the loops positions and returns the 
                    outputs self.loops_genes and self.genes[locus].is_loop 
                    described below.
            window (Optional [int.]): window around the loop positions (ie
                    around the 2 bins) for the seeking of overlapping genes
                    (Default: 0). Are considered "loops" genes, all genes
                    overlaping any position between :
                    * loops_start_bin1 - window and loops_end_bin1 + window or 
                    * loops_start_bin2 - window and loops_end_bin2 + window

        Note:
            The data importation requires a loops.info file that contains 
            the column indices of each information in the data file and 
            some additional information, in the following order:
            * (required) [0] Condition, [1] Filename, [2] Startline, 
              [3] Separator, [4] Bin1, [5] Bin2, [6] Binsize (in b),
            * (optional) [7] Score, [8] Pvalue,[9] Qvalue
        
        Note:    
            loops.info and data file have to be in the /HiC/Loops/ directory
        
        Note:    
            The position (pos) is the bin number. If data are binned at 2kb,
            the bin with number 10 corresponds to data between 20000 and 22000.

        Warning:
            This method needs a genomic sequence and a genomic annotation.
            If no annotation is loaded, the load_annotation method with the 
            default "sequence.gff3" file is computed. If no sequence is loaded, 
            the load_seq method with de defualt "sequence.fasta" is computed.
            To use another annotation or sequence, please load them to your
            Transcriptome instance with the following commands before using this 
            method:
            >>> from GRATIOSA import Genome, HiC
            >>> HC = HiC.HiC("ecoli")
            >>> g = Genome.Genome(HC.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> g.load_seq(filename=chosen_file2
            >>> HC.genes = g.genes
            >>> HC.length = g.length

        Warning:
            Make sure that the version of the genome is the same as
            the version used in the HiC analysis workflow

        Example:
            >>> from GRATIOSA import HiC
            >>> hc = HiC.HiC("dickeya")
            >>> hc.load_loops_genes()
            >>> hc.loops_pos['WT2kb_bin2000b']['loops']
            [254000,254001,254002,254003,...]
            >>> hc.loops_genes['WT2kb_bin2000b_w0b']['loops']
            ['Dda3937_00221','Dda3937_04438','Dda3937_03673',...]
        """
        if not hasattr(self, "seq"):
            gen = Genome.Genome(self.name)
            gen.load_seq()
            self.length = gen.length

        if not hasattr(self, "genes"):
            gen = Genome.Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes

        if not hasattr(self, "loops_genes"):
            self.loops_genes = {}

        if not hasattr(self, "loops_pos"):
            self.loops_pos = {}

        if isinstance(cond, str):
            cond = [cond]

        if not hasattr(self, "loops_pos") or cond == ["all"]:
            self.load_hic_loops(cond)
        else:
            unloaded_cond = set(cond) - set(self.loops_pos.keys())
            if unloaded_cond:
                self.load_hic_loops(list(unloaded_cond))

        if cond == ['all']:
            cond = self.loops_pos.keys()
        for c in cond:
            print(f"Loading loops genes in {c}_w{window}b")
            loops_pos = self.loops_pos[c]["loops"]
            is_loop = [False] * self.length
            for pos in loops_pos:
                is_loop[pos] = True

            self.loops_genes[f"{c}_w{window}b"] = {
                "loops": [], "no_loops": []}
            for locus in self.genes.keys():
                self.genes[locus].add_is_loop(c, False)
                gene = self.genes[locus]
                for pos in np.arange(gene.left - window,
                                     gene.right + 1 + window):
                    if gene.right + 1 + window >= self.length:
                        pos = pos - self.length
                    if is_loop[pos]:
                        self.genes[locus].add_is_loop(c, True)
                if self.genes[locus].is_loop[c]:
                    self.loops_genes[f"{c}_w{window}b"]["loops"].append(
                        locus)
                else:
                    self.loops_genes[f"{c}_w{window}b"]["no_loops"].append(
                        locus)

    def load_borders_genes(self, cond="all", window=0):
        """
        load_hic_borders_genes imports a list of borders from a data file 
        (typically a borders.tsv data file obtained with Chromosight) 
        containing at least the border position (bin number) and determines 
        the list of genes overlapping one of the borders.

        Creates:
            * self.borders_genes (dict. of dict.) 
                    created only if per_genes = True. New attribute of the 
                    Genome instance. Dictionary containing one dictionary 
                    per condition listed in borders.info. The key of this 
                    dictionary contains the condition name and the chosen
                    window size (for example "cond_w100b" for a 100b window).
                    Each subdictionary has 2 keys: "borders" and "no_borders" 
                    and contains the corresponding lists of genes. For 
                    example, self.borders_genes[cond_w0b]["borders"] returns
                    the list of genes that overlap the borders positions.
            * self.borders_pos (dict. of dict) 
                    new attribute of the HiC instance, dictionary containing 
                    one subdictionary per condition, with the shape 
                    self.borders_pos[cond] =
                    {'borders':list of positions that are in a border,
                    'no_borders':list of positions that are not in a border}
            * self.genes[locus].is_border (dict.) 
                    created only if per_genes = True.
                    New attribute of Gene instances related to the Genome
                    instance given as argument.Dictionary of shape
                    {condition:boolean}. The boolean is True if the gene
                    overlaps the border positions (window included).


        Args:
            cond (Optional [list of str.]): selection of one or several 
                    conditions (1 condition corresponds to 1 data file and 
                    has to be in the borders.info file). By default: 
                    cond ='all' ie all available data are loaded.
            per_genes(Optional [Bool.]): if True, determines the list of 
                    genes overlapping the borders positions and returns the 
                    outputs self.borders_genes and self.genes[locus].is_border 
                    described below.
            window (Optional [int.]): window around the border positions ie 
                    around the bin) for the seeking of overlapping genes
                    (Default: 0). All genes overlaping any position between 
                    border_start_bin - window and border_end_bin + window 
                    are considered "borders" genes.

        Note:
            The data importation requires a borders.info file that contains 
            the column indices of each information in the data file and some
            additional information, in the following order:
            * (required) [0] Condition, [1] Filename, [2] Startline,
              [3] Separator, [4] Bin, [5] Binsize (in b)
            * (optional) [6] Score, [7] Pvalue,[8] Qvalue
        
        Note:
            borders.info and data file have to be in the /HiC/Borders/ directory

        Note:
            The position (pos) is the bin number. If data are binned at 2kb,
            the bin with number 10 corresponds to data between 20000 and 22000.

        Warning:
            This method needs a genomic sequence and a genomic annotation.
            If no annotation is loaded, the load_annotation method with the 
            default "sequence.gff3" file is computed. If no sequence is loaded, 
            the load_seq method with de defualt "sequence.fasta" is computed.
            To use another annotation or sequence, please load them to your 
            Transcriptome instance with the following commands before using this 
            method:
            >>> from GRATIOSA import Genome, HiC
            >>> HC = HiC.HiC("ecoli")
            >>> g = Genome.Genome(HC.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> g.load_seq(filename=chosen_file2
            >>> HC.genes = g.genes
            >>> HC.length = g.length


        Warning:
            make sure that the version of the genome is the same as
            the version used in the HiC analysis workflow

        Example:
            >>> from GRATIOSA import HiC
            >>> hc = HiC.HiC("dickeya")
            >>> hc.load_borders_genes()
            >>> hc.borders_pos["WT2_2kb_bin2000b"]['borders']
            [20000,20001,20002,20003,20004,...]
            >>> hc.borders_genes['acid1kb_bin1000b_w0b']['borders']
            ['Dda3937_00158','Dda3937_01107','Dda3937_01108',...]
            >>> hc.genes['Dda3937_00221'].is_border
            {'acid1kb_bin1000b': False,'WT2_2kb_bin2000b': False}
        """

        if not hasattr(self, "seq"):
            gen = Genome.Genome(self.name)
            gen.load_seq()
            self.length = gen.length

        if not hasattr(self, "genes"):
            gen = Genome.Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes

        if not hasattr(self, "borders_genes"):
            self.borders_genes = {}

        if isinstance(cond, str):
            cond = [cond]

        if not hasattr(self, "borders_pos") or cond == ["all"]:
            self.load_hic_borders(cond)
        else:
            unloaded_cond = set(cond) - set(self.borders_pos.keys())
            if unloaded_cond:
                self.load_hic_borders(list(unloaded_cond))

        if cond == ["all"]:
            cond = self.borders_pos.keys()

        for c in cond:
            print(f"Loading borders genes in {c}_w{window}b")
            borders_pos = self.borders_pos[c]["borders"]
            is_border = [False] * self.length
            for pos in borders_pos:
                is_border[pos] = True

            self.borders_genes[f"{c}_w{window}b"] = {
                "borders": [], "no_borders": []}
            for locus in self.genes.keys():
                self.genes[locus].add_is_border(c, False)
                gene = self.genes[locus]
                for pos in np.arange(gene.left - window,
                                     gene.right + 1 + window):
                    if gene.right + 1 + window >= self.length:
                        pos = pos - self.length
                    if is_border[pos]:
                        self.genes[locus].add_is_border(c, True)
                if self.genes[locus].is_border[c]:
                    self.borders_genes[f"{c}_w{window}b"]["borders"].append(
                        locus)
                else:
                    self.borders_genes[f"{c}_w{window}b"]["no_borders"].append(
                        locus)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="GRATIOSA",
    version="0.0.2",
    author="Pineau Maïwenn, Forquet Raphaël, Meyer Sam",
    author_email="sam.meyer@insa-lyon.fr",
    description=" ",
    long_description=long_description,
    url="https://github.com/sammeyer2017/GRATIOSA.git",
    packages=['GRATIOSA'],
    install_requires=[
    	'numpy',
        'matplotlib',
        'pandas',
        'pysam',
        'scipy',
        'statsmodels'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
)

print(setuptools.find_packages())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Functions called by Transcriptome methods
"""
import numpy as np
from GRATIOSA.globvar import *
from GRATIOSA import Genome
import os
import pysam
import math

def add_expression_to_genes(genes_dict, 
                            cond, 
                            filename, 
                            tag_col, 
                            expr_col, 
                            is_log, 
                            separator):
    """
    Called by load_expression, adds expression data to Gene objects by 
    parsing a file with one column containing the locus tags of the genes and 
    one containing the expression data. 

    Creates a new attribute "expression" of the Gene objects contained in 
    the genes_dict given as argument. This attribute is a dictionary of shape 
    {condition: expression level [float.]}

    Args:
        genes_dict (dict.): dictionary of shape {locus: Gene object}
        cond (str.): condition name
        filename (str.): path to the file containing the expression data
        tag_col (int.): index of the column containing the locus tags in the 
                file
        expr_col (int.): index of the column containing the expression data
        is_log (bool.): True if the data are already in log2, such as log2RPKM.
                        False if the data are RPKM.
        separartor (str.): file separator

    Returns:
        Dictionary: Dict. of shape {'locus':Gene object} with in key the locus 
                tags associated to an expression data and in value the gene 
                object associated to each of these locus tags.
    Note: 
        Column numbering starts at 0.
    """
    genes_valid = {}
    except_locus = []
    expr0 = 0
    if separator == '\\t':
        separator = '\t'
    with open(filename, 'r') as f:
        header = next(f)
        header = header.strip().split(separator)
        header = header[expr_col:]
        genes_valid["conditions"] = header
        for line in f:
            line = line.strip().split(separator)
            try:
                if is_log == 'no':
                    genes_dict[line[tag_col]].add_expression(
                        cond, math.log(float(line[expr_col]), 2))
                elif is_log == 'yes':
                    genes_dict[line[tag_col]].add_expression(
                        cond, float(line[expr_col]))
                else:
                    print("Please set 'is_log' to 'yes' or 'no' in the expression.info file")
                genes_valid[line[tag_col]] = genes_dict[line[tag_col]]

            except KeyError:
                if line[tag_col] == 'none':
                    print("expressions without locus tag")
                else:
                    except_locus.append(line[tag_col])
            except Exception as e:
                if int(float(line[expr_col])) == 0:
                    expr0 += 1
                else:
                    print(e)

    print(f"{len(genes_valid.keys())} expressions were successfully loaded")
    if except_locus:
        if len(except_locus) > 20:
            print(f"{len(except_locus)} locus are not in annotation")
        else:
            print(f"{except_locus} are not in annotation")

    if expr0:
        print(f"{expr0} expressions are null and led to math domain error")
    return genes_valid


def load_fc_pval_cond(genes_dict, 
                      filename, 
                      condition,
                      tag_col, 
                      fc_col, 
                      separator, 
                      start_line, 
                      p_val_col=None):
    '''
    Called by load_fc_pval, allows expression data to be loaded by specifying
    a file containing the following information (one column for each type
    of information): locus tags of the genes, log2(Fold-Changes) and p-value.

    Args:
        genes_dict (dict.): dictionary of shape {locus: Gene object}
        filename (str.): path to the file containing the expression data
        cond (str.): condition name
        tag_col (int.): index of the column containing the locus tags in the 
                file
        fc_col (int.): index of the column containing the log2FC
        separartor (str.): file separator
        start_line (int.): file start line
        p_val_col (int.): index of the column containing the p-values in the
                file. If no index is indicated (default: None), the p-values 
                will be assigned to 0 for all genes.
    
    Returns:
        List: locus tags of all genes having a valid log2FC

    Note: 
        Column numbering starts at 0.
    '''
    genes_valid = []
    if separator == '\\t':
        separator = '\t'
    except_locus = []
    except_fc = 0
    with open(filename, 'r') as f:
        i = 0
        while i < start_line:
            header = next(f)
            i += 1
        for line in f:
            line = line.strip('\n').split(separator)
            try:
                if p_val_col:
                    genes_dict[line[tag_col]].add_fc_pval_cond(
                        float(line[fc_col]), condition, float(line[p_val_col]))
                else:
                    genes_dict[line[tag_col]].add_fc_pval_cond(
                        float(line[fc_col]), condition, float(0))
                genes_valid.append(line[tag_col])
            except BaseException:
                if line[tag_col] not in genes_dict.keys():
                    if line[tag_col] != '':
                        except_locus.append(line[tag_col])
                    else:
                        except_fc += 1
    f.close()
    if except_locus:
        if len(except_locus) > 20:
            print(f"\t{len(except_locus)} locus are not in annotation")
        else:
            print(f"\t{except_locus} are not in annotation")
    if except_fc:
        print(f"\t{except_fc} fc without locus")
    return genes_valid


def process_bam_paired_end(tr):
    """
    Called by load_cov_start_end and load_rnaseq_cov, process_bam_paired_end
    converts paired-end .bam files in .npy files containing the fragments.
    These files will enable a faster data loading for the next data
    importation with load_cov_start_end or load_rnaseq_cov.
    
    Creates:
        * For each condition listed in the bam_files.info file, this function
          creates a npz file containing the coverage data on both DNA strands
          (Rpos.npy and Rneg.npy). These files are saved in the /rnaseq_reads/
          directory
        *  A reads.info file containing the information required to associate
           each .npz  file to a condition name is also created (or completed if
           this file already exists) in the /rnaseq_reads/ directory. This file
           contains the following columns:
           [0] Condition [1] Reads filename [2] BAM filename

    Args:
        tr: Transcriptome instance

    Note: 
        The paired-end .bam reads files are, with a bam_files.info file, in the
        /rnaseq_reads/ directory. The bam_files.info file contains the following
        information:  [0] Condition [1] Reads filename for this condition
    
    Warnings: 
        Only for paired-end reads !
    """

    exist_cond = []
    path2dir = f"{basedir}data/{tr.name}/rnaseq_reads/"
    if not os.path.exists(path2dir + 'reads.info'):
        file = open(path2dir + 'reads.info', 'w')
        file.write('Condition\tReads filename\tBAM filename\n')
    else:
        file = open(path2dir + 'reads.info', 'r')
        header = next(file)
        for line in file:
            exist_cond.append(line.split("\t")[0])
    file.close()
    with open(path2dir + "bam_files.info", "r") as f:
        header = next(f)
        for line in f:  # for each condition
            line = line.strip().split('\t')
            bam_cond = line[0]
            if bam_cond not in exist_cond:
                bam_file = line[1]
                print(f"Converting .bam file for: {bam_cond}")
                if not os.path.exists(f"{path2dir}{bam_cond}.bai"):  
                        # if index needed, created using samtools (.bai file)
                    os.system("samtools index -b %s" % path2dir + bam_file)
                # BAM opening, alignment file object
                bamfile = pysam.AlignmentFile(path2dir + bam_file, "rb")

                # /!\ 0-based coordinate system /!\
                # lists storing coordinates of paired-end fragments for +/-
                # strands
                Rpos_start, Rpos_end, Rneg_start, Rneg_end = [], [], [], []

                for read in bamfile.fetch():
                    # if correctly paired and R1/R2 in different strands and
                    # fragment size < 1000 kb
                    if read.is_read1 and read.is_paired and not read.mate_is_unmapped and read.is_reverse != read.mate_is_reverse and abs(
                            read.template_length) < 1000:
                        if read.is_reverse:
                            Rneg_start.append(read.reference_end)
                            Rneg_end.append(
                                read.reference_end + abs(read.template_length))
                        elif not read.is_reverse:
                            Rpos_start.append(read.reference_start)
                            Rpos_end.append(
                                read.reference_start + read.template_length)
                bamfile.close()

                # conversion step into numpy array
                Rpos_start = np.array(Rpos_start, dtype=int)
                Rpos_end = np.array(Rpos_end, dtype=int)
                Rneg_start = np.array(Rneg_start, dtype=int)
                Rneg_end = np.array(Rneg_end, dtype=int)

                Rpos = np.column_stack((Rpos_start, Rpos_end))
                Rneg = np.column_stack((Rneg_start, Rneg_end))

                # delete rows where one coordinate is missing
                Rpos = Rpos[np.isfinite(Rpos).all(axis=1)]
                Rneg = Rneg[np.isfinite(Rneg).all(axis=1)]

                # save results ; .npz contains two .npy Rpos and Rneg
                file = open(path2dir + 'reads.info', 'a')
                file.write(f"{bam_cond}\t{bam_cond}_reads.npz\t{bam_file}\n")
                file.close()
                np.savez(f"{path2dir}{bam_cond}_reads.npz", 
                         Rpos=Rpos, 
                         Rneg=Rneg)


def cov_from_reads(tr):
    """
    Called by load_rnaseq_cov, cov_from_reads computes the coverage from 
    paired-end reads previously converted to .npz format (containing one .npy 
    per strand: Rpos.npy and Rneg.npy) with the process_bam_paired_end 
    function. 

    Creates:
        * For each condition listed in the reads.info file, this function
          creates a npz file containing the coverage data on both DNA strands:
          cov_pos.npy and cov_neg.npy. These files are saved in the 
          /rnaseq_cov/ directory and will enable a faster data loading for the 
          next data importation with load_rnaseq_cov.

        * A cov.info file containing the information required to associate
          each .npz  file to a condition name is also created (or completed if
          this file already exists) in the /rnaseq_cov/ directory. This file
          contains the following columns:  [0] Condition [1] Coverage filename

    Args:
        tr: Transcriptome instance

    Note: 
        The .npz files have to be, with a reads.info file (also created 
        by the process_bam_paired_end function), in the /rnaseq_reads/ directory. 
        This reads.info file contains at least the following information: 
        [0] Condition [1] Reads filename
    """
    # load npz file
    gen = Genome.Genome(tr.name)
    if not hasattr(gen, "seq"):
        gen.load_seq()
    genome_length = gen.length
    exist_cond = []
    path2cov = f"{basedir}data/{gen.name}/rnaseq_cov/"
    path2reads = f"{basedir}data/{gen.name}/rnaseq_reads/"
    if not os.path.exists(path2cov + 'cov.info'):
        file = open(path2cov + 'cov.info', 'w')
        file.write('Condition\tCoverage filename\n')
    else:
        file = open(path2cov + 'cov.info', 'r')
        header = next(file)
        for line in file:
            exist_cond.append(line.split("\t")[0])
    file.close()
    with open(path2reads + "reads.info", "r") as f:
        header = next(f)
        for line in f:  # for each condition
            line = line.strip().split('\t')

            if line[0] not in exist_cond:
                print('Converting reads in coverage for:', line[0])
                # load npz file corresponding to condition
                npzfile = np.load(path2reads + line[1])
                Rpos = npzfile["Rpos"]
                Rneg = npzfile["Rneg"]
                # init cov
                cov_pos = np.zeros(genome_length, dtype=int)
                cov_neg = np.zeros(genome_length, dtype=int)
                # compute cov on positive strand
                for start, end in Rpos:
                    cov_pos[start:end + 1] += 1
                # on negative strand
                for start, end in Rneg:
                    cov_neg[start:end + 1] += 1

                # save results
                file = open(path2cov + 'cov.info', 'a')
                file.write(f"{line[0]}\t{line[0]}_cov.npz\n")
                file.close()
                # save results ; .npz contains two .npy cov_pos and cov_neg
                np.savez(f"{path2cov}{line[0]}_cov.npz", 
                         cov_pos=cov_pos,
                         cov_neg=cov_neg)


def cov_start_stop_from_reads(tr):
    """
    Called by load_cov_start_end, cov_start_stop_from_reads computes the 
    density of RNA fragment starts and ends from paired-end reads previously 
    converted to .npz format (containing one .npy per strand: Rpos.npy and 
    Rneg.npy) with the process_bam_paired_end function.
    
    Creates:
        * For each condition listed in the reads.info file, this function
          creates a npz file containing the density of RNA fragment starts and
          ends on both DNA strands: cov_start_pos.npy, cov_start_neg.npy,
          cov_end_pos.npy and cov_end_neg.npy.
          These files are saved in the /rnaseq_cov/ directory and will enable
          a faster data loading for the next data importation with the
          cov_start_stop_from_reads function.
        * A cov_start_stop.info file containing the information required to 
          associate each .npz  file to a condition name is also created (or 
          completed if this file already exists) in the /rnaseq_cov/ directory. 
          This file contains the following columns:  
          [0] Condition [1] Coverage filename

    Args:
        tr: Transcriptome instance

    Note: 
        The .npz files have to be, with a reads.info file (also created by 
        process_bam_paired_end function), in the /rnaseq_reads/ directory. 
        This reads.info file contains at least the following information:
        [0] Condition [1] Reads filename
    """
    gen = Genome.Genome(tr.name)
    if not hasattr(gen, "seq"):
        gen.load_seq()
    genome_length = gen.length
    exist_cond = []
    path2cov = f"{basedir}data/{tr.name}/rnaseq_cov/"
    path2reads = f"{basedir}data/{tr.name}/rnaseq_reads/"
    if not os.path.exists(path2cov + 'cov_start_stop.info'):
        file = open(path2cov + 'cov_start_stop.info', 'w')
        file.write('Condition\tCov file\n')
    else:
        with open(path2cov + 'cov_start_stop.info', 'r') as file:
            header = next(file)
            for line in file:
                exist_cond.append(line[0])
    file.close()
    with open(path2reads + "reads.info", "r") as f:
        header = next(f)
        for line in f:  # for each condition
            line = line.strip().split('\t')
            if line[0] not in exist_cond:
                print('Searching for start and stop from reads in:', line[0])
                # load npz file corresponding to condition
                npzfile = np.load(path2reads + line[1])
                Rpos = npzfile["Rpos"]
                Rneg = npzfile["Rneg"]
                # init cov
                cov_start = {0: np.zeros(genome_length, dtype=int), 
                             1: np.zeros(genome_length, dtype=int)}
                cov_end = {0: np.zeros(genome_length, dtype=int), 
                           1: np.zeros(genome_length, dtype=int)}
                # compute cov
                # on positive strand
                for start, end in Rpos:
                    try:
                        cov_start[1][start - 1] += 1
                        cov_end[1][end - 1] += 1
                    except BaseException:
                        pass
                # on negative strand
                for start, end in Rneg:
                    try:
                        cov_start[0][start - 1] += 1
                        cov_end[0][end - 1] += 1
                    except BaseException:
                        pass
                # save results
                file = open(path2cov + 'cov_start_stop.info', 'a')
                fname = line[0] + "_cov_start_end"
                file.write(f'{line[0]}\t{fname}.npz\n')
                file.close()
                np.savez(f"{path2cov}{fname}.npz", 
                         cov_start_pos=cov_start[1],
                         cov_end_pos=cov_end[1],
                         cov_start_neg=cov_start[0],
                         cov_end_neg=cov_end[0])



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import numpy as np
import pandas as pd
from pathlib import Path
from GRATIOSA.useful_functions_Chipseq import *
from GRATIOSA.globvar import *
from GRATIOSA.Genome import Genome

class Chipseq:
    '''
    The Chipseq class is mainly used to collect coverage data (Chipseq signals) 
    along a genome. The associated methods and functions allow the binning and 
    averaging of these signals. The enrichment peaks positions can also be loaded 
    as an attribute of this class. 

    Each Chipseq instance has to be initialized with an organism name
    Example:
        >>> from GRATIOSA import Chipseq
        >>> ch = Chipseq.Chipseq("ecoli")
    '''

    def __init__(self, name):
        self.name = name


    def load_signal(self, cond="all"):
        '''
        load_signal imports a 1D distribution along the chromosome (typically a 
        CHIPSeq distribution) from a data file (typically a .bedGraph obtained 
        with bamCompare) containing: (1) bin starting positions, (2) bin ending 
        positions and (3) signal in  each bin.

        Creates a new attribute of the Chipseq instance: 
            * self.signal (dict.)
                   Dictionary of shape {condition: array containing one signal
                   value per genomic position}

        Args:
            cond (list of str.): 
                    Selection of one or several conditions (1 condition 
                    corresponds to 1 data file). By default: cond ='all' ie all 
                    available signals are loaded. All selected conditions have to 
                    be listed in signal.info file.

        Note:
            The data importation requires an signal.info file that contains the 
            columns positions of each information in the data file, in the following 
            order:
            [0] condition, [1] filename, [2] separator used in the data file, 
            [3] bin_start, [4] bin_end, [5] signal

        Example:
            >>> from GRATIOSA import Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> ch.load_signal()
            >>> ch.signal['Signal_Test']
            array([100,   0,   0, ..., 100, 100,   0])
        '''
        # gets the path to data and .info files
        path2dir = f"{basedir}data/{self.name}/chipseq/signals/"

        # converts from str to list (useful if only 1 condition is selected)
        if isinstance(cond, str):
            cond = [cond]

        # tries to open .info file
        list_cond_info = []  
        if Path(f"{path2dir}signals.info").exists():
            with open(f"{path2dir}signals.info", "r") as f:
                header = next(f)

                # 1 line in the .info file corresponds to 1 condition
                # (1 data file)
                for line in f:
                    line = line.strip().split('\t')
                    list_cond_info.append(line[0])

                    # For each selected condition:
                    if cond == ["all"] or line[0] in cond:
                        print('Loading condition', line[0])
                        # loads data
                        separator = line[2]
                        if separator == '\\t':
                            separator = '\t'
                        data = pd.read_csv(
                            path2dir + line[1], sep=separator, header=None)
                        signal = data.iloc[:, int(line[5])]

                        # computes binsize of each bin in the file
                        bs = data.iloc[:, int(line[4])] - \
                            data.iloc[:, int(line[3])]

                        # creates a new attribute to the Chipseq instance:
                        # signal[cond]= signal per pos
                        if not hasattr(self, "signal"):
                            self.signal = {}
                        self.signal[line[0]] = np.array(np.repeat(signal, bs))
            f.close()

        else:
            print(f"Unable to locate signals.info in {path2dir}")

        # warns that some selected conditions aren't described in the
        # .info file
        for c in cond:
            if c not in list_cond_info and c != "all":
                print(f"Please add {c} in signals.info")


    def load_binned_signal(self, binsize, cond="all"):
        '''
        If a file containing the data for the chosen condition and binsize 
        exists, load_binned_signal loads these data using the load_signal method
        Else, load_binned signal performs the following steps: 
        1 - imports a 1D distribution along the chromosome (typically a CHIPSeq 
        distribution) using the load_signal method
        2 - performs the binning at the chosen binsize using the 
        useful_functions_Chipseq.binning
        3 - saves the binned data in a file 

        Creates or adds items to 2 Chipseq instance attributes:
            * self.signal (dict.)
                    Dictionary of shape {condition: array containing one signal
                    value per genomic position (before binning)}
            * self.binned_signal (dict.)
                    Dictionary of shape {cond_bin: array containing one binned 
                    signal value per genomic position} with cond_bin the condition 
                    name merged with the bin size (example: WT_bin200b). 

        Args:
            binsize (int.)
            cond (list of str.): 
                    Selection of one or several conditions (1 condition 
                    corresponds to 1 data file). By default: cond ='all' ie all 
                    available signals are loaded. All selected conditions have 
                    to be listed in signal.info file.

        Note:
            See ChipSeq.load_signal method for the data requirements.

        Example:
            >>> from GRATIOSA import Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> ch.load_binned_signal(binsize=100,cond='Signal_Test')
            >>> ch.binned_signal["Signal_Test_bin100b"]
            array([100,   100,   100, ..., 10, 10, 10])
        '''
        if not hasattr(self, "length"):
            gen = Genome(self.name)
            gen.load_seq()
            self.length = gen.length

        # gets the path to data and .info files
        path2files = f"{basedir}data/{self.name}/chipseq/signals/"
        f_path = f"{path2files}/binned_data/"

        # gets the list of all existing conditions using .info file
        if Path(path2files + 'signals.info').exists():
            if cond == "all":
                with open(path2files + 'signals.info', "r") as fi:
                    header = next(fi)
                    cond = []
                    for line in fi:  # for each condition
                        line = line.strip().split('\t')
                        cond.append(line[0])
                fi.close()

            # converts cond from str to list if only 1 condition is selected
            elif isinstance(cond, str):
                cond = [cond]

        else:
            print(f"Unable to locate signals.info in {path2files}")

        # For each selected condition:
        if not hasattr(self, "binned_signal"):
            self.binned_signal = {}  # creates the new attribute

        for c in cond:
            cond_name = f"{c}_bin{str(binsize)}b"
            print('Condition:' + cond_name)

            # if a file contains data for this condition and binsize
            if Path(f"{f_path}{cond_name}.npy").exists():
                # adds the data to the binned_signal attribute
                print(f"loading existing file: {cond_name}.npy")
                binned_data = np.load(f"{f_path}{cond_name}.npy")
                self.binned_signal[cond_name] = np.array(
                    np.repeat(binned_data, binsize))[:self.length]

            # if no file contains data for this condition and binsize
            else:
                print("performing the binning")
                # loads data
                self.load_signal(cond=c)
                # performs the binning using the
                # useful_functions_Chipseq.binning function
                binned_data = binning(self.signal[c], binsize=binsize,
                                      stat="mean").statistic
                # adds the data to the binned_signal attribute
                self.binned_signal[cond_name] = np.array(
                    np.repeat(binned_data, binsize))[:self.length]

                # creates the "binned_data" directory if necessary
                Path(f_path).mkdir(parents=True, exist_ok=True)
                # saves the data as a .npy file
                np.save(f"{f_path}{cond_name}.npy", binned_data)

    def load_smoothed_signal(self, window, cond="all", *args, **kwargs):
        '''
        If a data file containing the data for the chosen condition and 
        smoothing exists, load_binned_signal loads these data using the 
        load_signal method
        Else, load_smoothed_signal performs the following steps: 
        1 - imports a 1D distribution along the chromosome (typically 
        a CHIPSeq distribution) using the load_signal method
        2 - performs the smoothing (moving average with the chosen 
        window size) using the useful_functions_Chipseq.smoothing
        3 - saves the smoothed data in a file 
        
        Creates or adds items to 2 Chipseq instance attributes:
            * self.signal (dict.)
                   Dictionary of shape {condition: array containing one signal
                   value per genomic position (before smoothing)}
            * self.smoothed_signal (dict.)
                    Dictionary of shape {cond_smoo: array containing one binned 
                    signal value per genomic position} with cond_smoo the 
                    condition name merged with the smoothing window size 
                    (example:  WT_smooth200b).


        Args:
            window (int.): window size. The value of the smoothed signal 
                    of a position p is equal to the average of the signal 
                    between p - window/2 and p + window/2.
            cond (list of str.): 
                    Selection of one or several conditions (1 condition 
                    corresponds to 1 data file). By default: cond ='all' ie 
                    all available signals are loaded. All selected conditions 
                    have to be listed in signal.info file.
        
        Note:
            See Chipseq.load_signal method for the data requirements.

        Example:
            >>> from GRATIOSA import Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> ch.load_smoothed_signal(window=100,cond='Signal_Test')
            >>> ch.smoothed_signal["Signal_Test_smooth100b"]
            array([100.1,   99.8,   98.8, ..., 10.1, 11.1, 10.8])
        '''
        # gets the path to data and .info files
        path2files = f"{basedir}data/{self.name}/chipseq/signals/"
        f_path = f"{path2files}/smoothed_data/"

        # gets the list of all existing conditions using .info file
        info_filename = kwargs.get('info_filename', 'signals.info')
        if Path(path2files + info_filename).exists():
            if cond == "all":
                with open(path2files + info_filename, "r") as fi:
                    header = next(fi)
                    cond = []
                    for line in fi:  # for each condition
                        line = line.strip().split('\t')
                        cond.append(line[0])
                fi.close()
            # converts cond from str to list if only 1 condition is selected
            elif isinstance(cond, str):
                cond = [cond]
        else:
            print("Unable to locate " + info_filename)

        if not hasattr(self, "smoothed_signal"):
            self.smoothed_signal = {}  # creates the new attribute

        # For each selected condition:
        for c in cond:
            cond_name = f"{c}_smooth{str(window)}b"
            print('Condition:' + cond_name)

            # if a file contains data for this condition and window size
            if Path(f"{f_path}{cond_name}.npy").exists():
                # adds the data to the smoothed_signal attribute
                print(f"loading existing file: {cond_name}.npy")
                self.smoothed_signal[cond_name] = np.load(
                    f"{f_path}{cond_name}.npy")

            # if no file contains data for this condition and binsize
            else:
                print("performing the smoothing")
                # loads data
                self.load_signal(cond=c)
                # performs the smoothing using the
                # useful_functions_Chipseq.smoothing function
                smooth_data = smoothing(
                    self.signal[c], window=window)
                # adds the data to the smoothed_signal attribute
                self.smoothed_signal[cond_name] = smooth_data
                # creates the "smoothed_data" directory if necessary
                Path(f_path).mkdir(parents=True, exist_ok=True)
                # saves the data as a .npy file
                np.save(f"{f_path}{cond_name}.npy", smooth_data)


    def load_signals_average(self, list_cond, average_name, *args, **kwargs):
        '''
        Load_signals_average computes and loads the average of signals replicates.
        First, the function loads the signals (which must be listed in the 
        signals.info file in the /chipseq/signals/ folder) using the load_signal 
        method. It can then process the data using the load_binned_signal or 
        load_smoothed_signal methods. Finally, the average of these signals at 
        each genomic position of the genome is calculated. This average signal 
        is assigned to the Chipseq instance as signals_average attribute:
        * self.signals_average (dict.):
                Dictionary of shape {average_name: array containing one 
                averaged signal value per genomic position}. 

        Args:
            list_cond (list of str.): 
                    Selection of conditions that will be averaged. All selected 
                    conditions have to be listed in signal.info file.
            average_name (str.): 
                    Name of the obtained signal
            data_treatment (Optionnal [str.] "binning", "smoothing" or None):
                    Treatment to be applied to the different signals 
                    before averaging them (None by default).
            window (Optionnal [int.]): 
                    window size used only if data_treatment = "smoothing"
                    The value of the smoothed signal of a position p is 
                    equal to the average of the signal between 
                    p - window/2 and p + window/2
            binsize (Optionnal [int.]): bin size used only if data_treatment = "binning"
        
        Note:
            See Chipseq.load_signal method for the data requirements

        Example:
            >>> from GRATIOSA import Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> ch.load_signals_average(list_cond=["Signal1","Signal2"],
            ...                         average_name="Mean_signal",
            ...                         data_treatment = "smoothing",
            ...                         window=500)
            >>> ch.signals_average["Mean_signal"]
            array([100.1,   99.8,   98.8, ..., 10.1, 11.1, 10.8])
        '''

        f_path = f"{basedir}data/{self.name}/chipseq/signals/average_data/"
        data_treatment = kwargs.get('data_treatment', None)
        if not hasattr(self, "signals_average"):
            self.signals_average = {}
        # test if a file for this condition (cond_name) and this binsize
        # already exists

        if Path(f"{f_path}{average_name}.npy").exists():
            with open(f"{f_path}data.info", "r") as f:
                header = next(f).strip("\n").split("\t")
                # 1 line in the .info file corresponds to 1 condition
                for line in f:
                    line = line.strip().split('\t')
                    if line[0] == average_name:
                        print("loading the file obtained with the following parameters:\n")
                        for i in np.arange(len(header)):
                            print(f"{header[i]}: {line[i]}\n")

                        print("Please change 'average_name' to use other signals")
            self.signals_average[average_name] = np.load(
                f"{f_path}{average_name}.npy")

        else:
            print("performing the average")

            if data_treatment == "binning":
                print("data_treatment: binning")
                try:
                    size = kwargs.get('binsize')
                except:
                    sys.exit("please give a binsize")
                self.load_binned_signal(size, cond=list_cond)
                data_to_av = [
                    self.binned_signal[f"{i}_bin{str(size)}b"] for i in list_cond]

            elif data_treatment == "smoothing":
                print("data_treatment: smoothing")
                try:
                    size = kwargs.get('window')
                except:
                    sys.exit(
                        "please give a window size using \"window\" argument")
                self.load_smoothed_signal(size, cond=list_cond)
                data_to_av = [
                    self.smoothed_signal[f"{i}_smooth{str(size)}b"] for i in list_cond]
            else:
                data_treatment = "No treatment"
                size = "NA"
                print("data_treatment: No treatment")
                self.load_signal(cond=list_cond)
                data_to_av = [self.signal[i] for i in list_cond]

            # performs the average between replicates
            average = np.mean(data_to_av, axis=0)

            # adds the signals_average attribute to Chipseq instance
            self.signals_average[average_name] = average

            # saves the data as a .npy file
            Path(f_path).mkdir(parents=True, exist_ok=True)
            np.save(f"{f_path}{average_name}.npy", average)

            # saves information about the used replicates and treatment
            if not Path(f"{f_path}data.info").exists():
                file = open(f"{f_path}data.info", 'w')
                file.write('Name\t Replicates\t Data treatment\t Size (window or bin, in b)')
                file.close()
            file = open(f"{f_path}data.info", 'a')
            file.write(f"\n{average_name}\t{list_cond}\t{data_treatment}\t{size}")
            file.close()


    def get_all_signals(self):
        '''
        get_all_signals groups all loaded signals into a single attribute:

            * self.all_signals (dict.): 
                   Dictionary of shape {condition: array containing one signal
                   value per genomic position}
        '''
        d = {}
        for attr in ("signal","binned_signal","smoothed_signal","signals_average"):
            if hasattr(self, attr):
                d.update(getattr(self, attr))
        self.all_signals = d


    def load_signal_per_genes(self, cond='all', window=0):
        """ 
        Load_signal_per_genes computes the mean signal for each Gene (i.e. 
        the mean signal between the gene start and gene end). Genomic signals
        have to be loaded (using load_signal,load_binned_signal, 
        load_smoothed_signal or load_signals_average methods) before using 
        load_signal_per_genes.

        Creates: 
            * self.genes[locus].signal (float.): 
                    new attribute of Gene instances 
                    related to the Chipseq instance given as argument.Contains 
                    the Gene mean signal. 
            * self.signals_gene (dict. of dict.): 
                    Dictionary containing one  subdictionary per condition given 
                    in input. Each  subdictionary contains the signal (value) for 
                    each gene (locus_tag as key). 

        Args:
            cond (Optional [list of str.]): selection of one or several 
                    conditions (1 condition corresponds to 1 data file).
                    By default cond ='all' ie all available loaded signals are 
                    used.
            window (Optional int.): to include the signal around the gene. 
                    The mean signal will be computed between
                    gene start - window and gene end + window.
                    By default, window = 0

        Warning:
            This method needs a genomic annotation. If no annotation is 
            loaded, the load_annotation method with the default "sequence.gff3" 
            file is computed. To use another annotation, please load an annotation 
            to your Transcriptome instance with the following commands before using 
            this method:
            >>> from GRATIOSA import Genome, Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> g = Genome.Genome(ch.name)
            >>> g.load_annotation(annot_file=chosen_file)
            >>> ch.genes = g.genes    

        Example:
            >>> from GRATIOSA import Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> ch.load_signal()
            >>> ch.load_signal_per_genes()
            >>> ch.signals_gene["WT"]["b0984"]
            0.40348639903919425
            >>> ch.genes["b0984"].signal
            {'WT': 0.40348639903919425,
             'Signal_Test': 0.210505235030067}
        """
        self.get_all_signals()
        if not hasattr(self, "signals_gene"):
            self.signals_gene = {}
        if not hasattr(self, "genes"):
            gen = Genome(self.name)
            gen.load_annotation()
            self.genes = gen.genes

        if not hasattr(self, "signals_gene"):
            self.signals_gene = {cond_name: {}}

        if cond == "all":
            cond = list(self.all_signals.keys())
        elif isinstance(cond, str):
            cond = [cond]

        for c in cond:
            self.signals_gene[c] = {}
            for locus in self.genes.keys():
                g = self.genes[locus]
                s = np.mean(self.all_signals[c][g.left-window:g.right+window])
                self.genes[locus].add_signal(c, s)
                self.signals_gene[c][locus] = s
        print(f"Gene signals successfully computed for: {cond}")


    def load_peaks(self):
        """ 
        load_peaks imports a list of peaks from a data file (typically 
        a .BED file of peaks obtained with MACS2) containing, at least,
        the start and end positions of peaks.

        Creates:
            * self.peaks (dict. of dict.) : new attribut of the Chipseq instance. 
                    dictionary containing one subdictionary per condition 
                    with the shape {(start,end):value}. The key of each
                    subdictionary is a tupe containing the start and end 
                    positions of one peak. The value can for example be the 
                    peak score or the peak height.

        Note:
            The data importation requires a peaks.info file that contains the
            column indices of each information in the data file and some 
            additional information, in the following order:
            [0] Condition, [1] Filename, [2] Startline, 
            [3] Separator, [4] StartCol, [5] StopCol
            [6] Peak value
            peaks.info and data file have to be in the /chipseq/peaks/ directory 

        Example:
            >>> from GRATIOSA import Chipseq
            >>> ch = Chipseq.Chipseq("ecoli")
            >>> ch.load_peaks()
            >>> ch.peaks['test']
            {(840081, 840400): 4.89872,
            (919419, 919785): 5.85158,
            (937220, 937483): 4.87632,
            ...}
        """

        # gets the path to data and .info files
        path2dir = f"{basedir}data/{self.name}/chipseq/peaks/"

        # tries to open peaks.info file
        if Path(f"{path2dir}peaks.info").exists():
            with open(f"{path2dir}peaks.info", "r") as f:
                skiphead = next(f)
                # 1 line in peaks.info corresponds to 1 condition (1 data file)
                for line in f:
                    line = line.strip('\n').split('\t')

                    # loads data file information for this condition
                    cond = line[0]
                    print(line[0])
                    path2file = f"{path2dir}{line[1]}"
                    startline, sep = int(line[2]), line[3]
                    start_col, end_col = int(line[4]), int(line[5])
                    val_col = None
                    if len(line) > 6:
                        try : 
                            val_col = int(line[6])
                        except :
                            pass
                    # creates the new attribute using the load_sites_cond function
                    # from useful_functions_Chipseq
                    if not hasattr(self, "peaks"):
                        self.peaks = {}  # creates the new attribute
                    self.peaks[cond] = load_sites_cond(path2file, 
                                                       startline, 
                                                       sep,
                                                       start_col, 
                                                       end_col,
                                                       val_col)
        else:
            print("No peaks.info, unable to load peaks")
